import click

from gen3_util.buckets.lister import ls
from gen3_util.repo import NaturalOrderGroup, CLIOutput
from gen3_util.config import Config, ensure_auth


@click.group(name='buckets', cls=NaturalOrderGroup, invoke_without_command=True)
@click.pass_context
def bucket_group(ctx: click.Context):
    """Project buckets."""
    if ctx.invoked_subcommand is None:
        ctx.invoke(ls_command)


@bucket_group.command(name="ls")
@click.pass_obj
def ls_command(config: Config):
    """List buckets managed by commons."""
    with CLIOutput(config=config) as output:
        try:
            auth = ensure_auth(config=config, validate=True)
            assert auth, "auth required"
            output.update({'endpoint': auth.endpoint})
            output.update(ls(config))
        except Exception as e:
            output.update({'msg': str(e)})
            output.exit_code = 1




import click
import pathlib
import sys

from halo import Halo
from gen3_tracker import Config, ENV_VARIABLE_PREFIX
from gen3_tracker.common import INFO_COLOR, ERROR_COLOR


@click.group()
@click.option('--project_id', default=None, show_default=True,
              help="Gen3 program-project", envvar=f"{ENV_VARIABLE_PREFIX}PROJECT_ID")
@click.pass_context
def meta(ctx, project_id):
    """Manage the META directory."""
    pass


@meta.command()
@click.option('--project_id', default=None, show_default=True,
              help="Gen3 program-project", envvar=f"{ENV_VARIABLE_PREFIX}PROJECT_ID")
@click.option('--debug', is_flag=True)
@click.pass_context
def init(ctx, project_id, debug):
    """Initialize the META directory based on the MANIFEST."""
    try:
        from gen3_tracker.common import INFO_COLOR, ERROR_COLOR
        from gen3_tracker.meta.skeleton import update_meta_files
        from gen3_tracker.meta.validator import validate as validate_dir

        with Halo(text='Generating', spinner='line', placement='right', color='white'):
            config: Config = ctx.obj
            if not project_id:
                project_id = config.gen3.project_id
            updated_files = update_meta_files(config.dry_run, project_id)
        click.secho(f"Updated {len(updated_files)} metadata files.", fg=INFO_COLOR, file=sys.stderr)
        result = validate_dir('META', project_id)
        click.secho(result, fg=INFO_COLOR, file=sys.stderr)

    except Exception as e:
        click.secho(str(e), fg=ERROR_COLOR, file=sys.stderr)
        if ctx.obj.debug or debug:
            raise
        exit(1)


@meta.command()
@click.argument('directory', type=click.Path(exists=True), default='META')
@click.option('--debug', is_flag=True)
@click.pass_obj
def validate(ctx, directory, debug):
    """Validate FHIR data"""
    try:
        from gen3_tracker.meta.validator import validate as validate_dir
        with Halo(text='Validating', spinner='line', placement='right', color='white'):
            result = validate_dir(directory, project_id=ctx.gen3.project_id)
        click.secho(result.resources, fg=INFO_COLOR, file=sys.stderr)
        for _ in result.exceptions:
            click.secho(f"{_.path}:{_.offset} {_.exception}", fg=ERROR_COLOR, file=sys.stderr)
        if result.exceptions:
            sys.exit(1)
    except Exception as e:
        click.secho(str(e), fg=ERROR_COLOR, file=sys.stderr)
        if debug or ctx.debug:
            raise


@meta.command("graph")
@click.argument("directory_path",
                type=click.Path(exists=True, file_okay=False),
                default="META", required=False)
@click.argument("output_path",
                type=click.Path(file_okay=True),
                default="meta.html", required=False)
@click.option('--browser', default=False, show_default=True, is_flag=True, help='Open the graph in a browser.')
@click.pass_obj
def render_graph(config: Config, directory_path: str, output_path: str, browser: bool):
    """Render metadata as a network graph.

    \b
    directory_path: The directory path to the metadata. default [META]
    output_path: The output path for the network graph. default [meta.html]
    """
    try:
        from gen3_tracker.meta.visualizer import create_network_graph
        import webbrowser

        assert pathlib.Path(directory_path).exists(), f"Directory {directory_path} does not exist."
        with Halo(text='Graphing', spinner='line', placement='right', color='white'):
            output_path = pathlib.Path(output_path)
            create_network_graph(directory_path, output_path)
            url = f"file://{output_path.absolute()}"
        click.secho(f"Saved {output_path}, open it in your browser to view the network.", fg=INFO_COLOR, file=sys.stderr)
        if browser:
            webbrowser.open(url)
    except Exception as e:
        click.secho(str(e), fg=ERROR_COLOR, file=sys.stderr)
        if config.debug:
            raise


@meta.command("dataframe")
@click.argument('data_type',
                required=True,
                type=click.Choice(['Specimen', 'DocumentReference', 'ResearchSubject', "MedicationAdministration", "GroupMember"]),
                default=None)
@click.argument("directory_path",
                type=click.Path(exists=True, file_okay=False),
                default="./META", required=False)
@click.argument("output_path",
                type=click.Path(file_okay=True), required=False)
@click.option('--dtale', 'launch_dtale', default=False, show_default=True, is_flag=True, help='Open the graph in a browser using the dtale package for interactive data exploration.')
@click.option('--debug', is_flag=True)
@click.pass_obj
def render_df(config: Config, directory_path: str, output_path: str, launch_dtale: bool, data_type: str, debug: bool):
    """Render a metadata dataframe.

    \b
    directory_path: The directory path to the metadata.
    output_path: The output path for the dataframe. default [meta.csv]
    """
    try:
        from gen3_tracker.meta.dataframer import create_dataframe
        df = create_dataframe(directory_path, config.work_dir, data_type)

        if launch_dtale:
            import dtale
            dtale.show(df, subprocess=False, open_browser=True, port=40000)
        else:
            # export to csv
            file_name = output_path if output_path else f"{data_type}.csv"
            df.to_csv(file_name, index=False)
            click.secho(f"Saved {file_name}", fg=INFO_COLOR, file=sys.stderr)
    except Exception as e:
        click.secho(str(e), fg=ERROR_COLOR, file=sys.stderr)
        if config.debug or debug:
            raise


meta.add_command(render_df, name='df')



import inflection

from pydantic import BaseModel, computed_field
from typing import Dict, List, Optional, Tuple


#######################
# FHIR HELPER METHODS #
#######################


def get_nested_value(d: dict, keys: list):
    for key in keys:
        try:
            d = d[key]
        except (KeyError, IndexError, TypeError):
            return None
    return d


def normalize_coding(resource_dict: Dict) -> List[Tuple[str, str]]:
    """normalize any nested coding"""

    def extract_coding(coding_list):
        # return a concatenated string
        # or alternatively return an array
        return [coding.get("display", coding.get("code", "")) for coding in coding_list]

    def find_codings_in_dict(
        d: dict, parent_key: str = ""
    ) -> list[tuple[str, str]]:  # TODO - parent_key not used?
        codings = []
        for key, value in d.items():
            if isinstance(value, list):
                for item in value:
                    if isinstance(item, dict):
                        # Check if the dict contains a 'coding' list
                        if "coding" in item and isinstance(item["coding"], list):
                            coding_string = extract_coding(item["coding"])
                            codings.append((coding_string, key))
                        if "code" in item:
                            coding_string = item.get("display", item.get("code"))
                            codings.append((coding_string, key))

                        # Recursively search in the dict
                        codings.extend(find_codings_in_dict(item, key))
            elif isinstance(value, dict):
                # Check if the dict contains a 'coding' list
                if "coding" in value and isinstance(value["coding"], list):
                    coding_string = extract_coding(value["coding"])
                    codings.append((coding_string, key))

                # Recursively search in the dict
                codings.extend(find_codings_in_dict(value, key))
        return codings

    return find_codings_in_dict(resource_dict)


def normalize_value(resource_dict: dict) -> tuple[Optional[str], Optional[str]]:
    """return a tuple containing the normalized value and the name of the field it was derived from"""

    if "valueQuantity" in resource_dict:
        value = resource_dict["valueQuantity"]
        value_normalized = f"{value['value']} {value.get('unit', '')}"
        value_source = "valueQuantity"
    elif "valueCodeableConcept" in resource_dict:
        value = resource_dict["valueCodeableConcept"]
        value_normalized = " ".join(
            [
                coding.get("display", coding.get("code", ""))
                for coding in value.get("coding", [])
            ]
        )
        value_source = "valueCodeableConcept"
    elif "valueCoding" in resource_dict:
        value = resource_dict["valueCoding"]
        value_normalized = value["display"]
        value_source = "valueCoding"
    elif "valueString" in resource_dict:
        value_normalized = resource_dict["valueString"]
        value_source = "valueString"
    elif "valueCode" in resource_dict:
        value_normalized = resource_dict["valueCode"]
        value_source = "valueCode"
    elif "valueBoolean" in resource_dict:
        value_normalized = str(resource_dict["valueBoolean"])
        value_source = "valueBoolean"
    elif "valueInteger" in resource_dict:
        value_normalized = str(resource_dict["valueInteger"])
        value_source = "valueInteger"
    elif "valueRange" in resource_dict:
        value = resource_dict["valueRange"]
        low = value["low"]
        high = value["high"]
        value_normalized = f"{low['value']} - {high['value']} {low.get('unit', '')}"
        value_source = "valueRange"
    elif "valueRatio" in resource_dict:
        value = resource_dict["valueRatio"]
        numerator = value["numerator"]
        denominator = value["denominator"]
        value_normalized = f"{numerator['value']} {numerator.get('unit', '')}/{denominator['value']} {denominator.get('unit', '')}"
        value_source = "valueRatio"
    elif "valueSampledData" in resource_dict:
        value = resource_dict["valueSampledData"]
        value_normalized = value["data"]
        value_source = "valueSampledData"
    elif "valueTime" in resource_dict:
        value_normalized = resource_dict["valueTime"]
        value_source = "valueTime"
    elif "valueDateTime" in resource_dict:
        value_normalized = resource_dict["valueDateTime"]
        value_source = "valueDateTime"
    elif "valuePeriod" in resource_dict:
        value = resource_dict["valuePeriod"]
        value_normalized = f"{value['start']} to {value['end']}"
        value_source = "valuePeriod"
    elif "valueUrl" in resource_dict:
        value_normalized = resource_dict["valueUrl"]
        value_source = "valueUrl"
    elif "valueDate" in resource_dict:
        value_normalized = resource_dict["valueDate"]
        value_source = "valueDate"
    elif "valueCount" in resource_dict:
        value_normalized = resource_dict["valueCount"]["value"]
        value_source = "valueCount"
    else:
        value_normalized, value_source = None, None
        # for debugging...
        # raise ValueError(f"value[x] not found in {resource_dict}")

    return value_normalized, value_source


def normalize_for_guppy(key: str):
    """normalize a key so that it can be loaded into Guppy as a column name"""
    guppy_table = str.maketrans(
        {
            ".": "",
            " ": "_",
            "[": "",
            "]": "",
            "'": "",
            ")": "",
            "(": "",
            ",": "",
            "/": "_per_",
            "-": "to",
            "#": "number",
            "+": "_plus_",
            "%": "percent",
            "&": "_and_",
        }
    )
    return key.translate(guppy_table)


def traverse(resource):
    """simplify a resource's fields, returned as a dict of values,
    where keys are prefixed with "resourceType_" """

    final_subject = {}
    simplified_subject = SimplifiedResource.build(resource=resource).simplified
    prefix = simplified_subject["resourceType"].lower()
    for k, v in simplified_subject.items():
        if k in ["resourceType"]:
            continue
        final_subject[f"{prefix}_{k}"] = v

    return final_subject


########################
# FHIR BUILDER OBJECTS #
########################


class SimplifiedFHIR(BaseModel):
    """All simplifiers should inherit from this class."""

    warnings: list[str] = []
    """A list of warnings generated during the simplification process."""
    resource: dict
    """The FHIR resource to be simplified."""

    @computed_field()
    @property
    def simplified(self) -> dict:
        _ = self.identifiers.copy() if self.identifiers else {}
        _.update(self.scalars)
        _.update(self.codings)
        _.update(self.extensions)
        _.update(self.values)
        return _

    def simplify_extensions(
        self, resource: dict = None, _extensions: dict = None
    ) -> dict:
        """Extract extension values, derive key from extension url"""

        def _populate_simplified_extension(extension: dict):
            # simple extension
            value_normalized, extension_key = normalize_value(extension)
            extension_key = extension["url"].split("/")[-1]
            extension_key = (
                inflection.underscore(extension_key)
                .removesuffix(".json")
                .removeprefix("structure_definition_")
            )
            assert (
                value_normalized is not None
            ), f"extension: {extension_key} = {value_normalized} {extension}"
            _extensions[extension_key] = value_normalized

        if not _extensions:
            _extensions = {}

        if not resource:
            resource = self.resource

        for _ in resource.get("extension", [resource]):
            # special case data looks like this skip it, no extension to extract
            if set(_.keys()) == {"url", "size", "hash", "title"}:
                continue
            elif "extension" not in _.keys():
                if "resourceType" not in _.keys():
                    _populate_simplified_extension(_)
                continue
            elif set(_.keys()) == {"url", "extension"}:
                for child_extension in _["extension"]:
                    self.simplify_extensions(
                        resource=child_extension, _extensions=_extensions
                    )

        return _extensions

    @computed_field
    @property
    def extensions(self) -> dict:
        return self.simplify_extensions()

    @computed_field
    @property
    def scalars(self) -> dict:
        """Return a dictionary of scalar values."""
        return {
            k: v
            for k, v in self.resource.items()
            if (not isinstance(v, list) and not isinstance(v, dict))
        }

    @computed_field
    @property
    def codings(self) -> dict:
        """Return a dictionary of scalar values."""
        _codings = {}
        for k, v in self.resource.items():
            # these are handled in separate methods
            if k in ["identifier", "extension", "component", "code"]:
                continue
            elif isinstance(v, list):
                for elem in v:
                    if isinstance(elem, dict):
                        # TODO: implement hierarchy of codes rather than just taking last code?
                        for value, source in normalize_coding(elem):
                            if len(v) > 1 and get_nested_value(
                                elem, [source, 0, "system"]
                            ):
                                _codings[elem[source][0]["system"].split("/")[-1]] = (
                                    value
                                )
                            else:
                                _codings[k] = value
            elif isinstance(v, dict):
                for value, elem in normalize_coding(v):
                    _codings[k] = value

        return _codings

    @computed_field
    @property
    def identifiers(self) -> dict:
        """Return the first of a resource and any other resources"""
        identifiers = self.resource.get("identifier", [])
        identifiers_len = len(identifiers)

        if not identifiers_len:
            return {"identifier": None}
        elif identifiers_len == 1:
            return {"identifier": identifiers[0].get("value")}
        else:
            # assume that 0th identifier is the base identifier
            base_identifier = {
                (
                    "identifier"
                    if  i == 0 or identifier.get("use", "") == "official"
                    else identifier.get("system").split("/")[-1]
                ): identifier.get("value")
                for i, identifier in enumerate(identifiers)
            }

            return base_identifier

    @computed_field
    @property
    def values(self) -> dict:
        """Return a dictionary of source:value."""
        # FIXME: values that are scalars are processed twice: once in scalars once here in values (eg valueString)
        value, source = normalize_value(self.resource)
        if not value:
            return {}

        # update the key if code information is available
        if self.resource.get("code", {}).get("text", None):
            source = self.resource["code"]["text"]
        return {source: value}


class SimplifiedObservation(SimplifiedFHIR):
    @computed_field
    @property
    def codings(self) -> dict:
        """does everything but gets rid of code since that's dealt with in values"""
        _codings = super().codings
        if "code" in _codings:
            del _codings["code"]

        return _codings

    # TODO: remove after data fix
    @computed_field
    @property
    def scalars(self) -> dict:
        """Return a dictionary of scalar values."""
        return {
            k: v
            for k, v in self.resource.items()
            if (
                not isinstance(v, list)
                and not isinstance(v, dict)
                and not k == "valueString"
            )
        }

    @computed_field
    @property
    def values(self) -> dict:
        """Return a dictionary of 'value':value or <component>:value.
        https://build.fhir.org/observation-definitions.html#Observation.component
        """

        # get top-level value in dict if it exists
        _values = super().values

        if len(_values) == 0:
            assert (
                "component" in self.resource
            ), "no component nor top-level value found"

            # get component codes
            if "component" in self.resource:
                for component in self.resource["component"]:
                    value, source = normalize_value(component)
                    if component.get("code", {}).get("text", None):
                        source = component["code"]["text"]
                    if not value:
                        continue
                    _values[source] = value

        # knowing there's now at least 1 item in _values
        if "component" in self.resource:
            # ensure no top-level value is not duplicating a component code value
            # TODO: ensure this value_key corresponds to percent_tumor on some runs due to getting display
            value_key = [k for k in _values][0]
            assert (
                value_key not in self.resource["component"]
            ), """duplicate code value found, only specify the code value in the component, see Rule obs-7
                https://build.fhir.org/observation.html#invs"""

            # get component codes
            if "component" in self.resource:
                for component in self.resource["component"]:
                    value, source = normalize_value(component)
                    if component.get("code", {}).get("text", None):
                        source = component["code"]["text"]
                    if not value:
                        continue
                    _values[source] = value
        if "code" in self.resource and "text" in self.resource["code"]:
            _values["observation_code"] = self.resource["code"]["text"]

        assert len(_values) > 0, f"no values found in Observation: {self.resource}"

        return _values


class SimplifiedDocumentReference(SimplifiedFHIR):
    @computed_field
    @property
    def values(self) -> dict:
        """Return a dictionary of 'value':value."""
        _values = super().values
        for content in self.resource.get("content", []):
            if "attachment" in content:
                for k, v in SimplifiedFHIR(
                    resource=content["attachment"]
                ).simplified.items():
                    if k in ["identifier", "extension"]:
                        continue
                    _values[k] = v
        return _values


class SimplifiedMedicationAdministration(SimplifiedFHIR):
    @computed_field
    @property
    def values(self) -> dict:
        """Return a dictionary of 'value':value."""
        _values = super().values
        # Plucking out fields that didn't get picked up by default class simplifier.'
        dose_value = self.resource.get("dosage", {}).get("dose", {}).get("value", None)
        if dose_value:
            _values["total_dosage"] = dose_value
        occurenceTiming = (
            self.resource.get("occurenceTiming", {})
            .get("repeat", {})
            .get("boundsRange")
        )
        if occurenceTiming:
            low = occurenceTiming.get("low", {}).get("value")
            _values["index_date_start_days"] = low if low else None
            high = occurenceTiming.get("high", {}).get("value")
            _values["index_date_end_days"] = high if high else None
        for notes in self.resource.get("note", []):
            note = notes.get("value", None)
            if note:
                # Probably best to concat notes together
                _values["notes"] = _values["notes"] + "; " + note
        for identifier in self.resource.get("identifier", []):
            system = identifier.get("system", None)
            if system:
                if system.split("/")[-1] == "regimen":
                    _values["regimen_id"] = identifier["value"]
        return _values


class SimplifiedCondition(SimplifiedFHIR):
    @computed_field
    @property
    def codings(self) -> dict:
        # only go through the work if code exists
        if "code" not in self.resource:
            return {}

        # get field name
        codings_dict = super().codings
        if "category" in codings_dict:
            key = codings_dict["category"]
            del codings_dict["category"]
        else:
            key = "code"

        # TODO: implement hierarchy of codes rather than just taking last code?
        value, _ = normalize_coding(self.resource["code"])[-1]
        return {key: value}


class SimplifiedGroup(SimplifiedFHIR):
    @computed_field
    @property
    def members(self) -> list:
        """ "Get the list of the members of the group"""

        members = []

        # for each member, add its uuid to the list
        for member_dict in self.resource["member"]:
            member_reference = member_dict["entity"].get("reference", None)
            if member_reference:
                members.append(member_reference.split("/")[-1])

        # return all uuids
        return members


class SimplifiedResource(object):
    """A simplified FHIR resource, a factory method."""

    @staticmethod
    def build(resource: dict) -> SimplifiedFHIR:
        """Return a simplified FHIR resource."""

        resource_type = resource.get("resourceType", None)
        if resource_type == "Observation":
            return SimplifiedObservation(resource=resource)
        if resource_type == "DocumentReference":
            return SimplifiedDocumentReference(resource=resource)
        if resource_type == "Condition":
            return SimplifiedCondition(resource=resource)
        if resource_type == "MedicationAdministration":
            return SimplifiedMedicationAdministration(resource=resource)
        if resource_type == "Group":
            return SimplifiedGroup(resource=resource)
        return SimplifiedFHIR(resource=resource)



from pydantic import BaseModel

from gen3_util.buckets import get_buckets
from gen3_util.config import Config, ensure_auth


class LogBuckets(BaseModel):
    endpoint: str
    """The commons url"""
    buckets: dict
    """List of buckets"""

# add a cache to this to avoid hitting arborist too much (or at all), time it out after 1 hour


def ls(config: Config):
    """List projects."""
    auth = ensure_auth(config=config)
    buckets = get_buckets(auth=auth)
    return LogBuckets(**{
        'endpoint': auth.endpoint,
        'buckets': buckets
    })



import logging
import os
import pathlib
import sys
from configparser import ConfigParser
from datetime import datetime, timezone, timedelta
from typing import Generator

import click
import jwt
import importlib.resources as pkg_resources
import requests
import yaml
from gen3.auth import Gen3Auth
from gen3.file import Gen3File
from gen3.index import Gen3Index

import gen3_tracker
from gen3_tracker import Config
from gen3_tracker.common import read_yaml, PROJECT_DIRECTORIES, PROJECT_DIR, PROJECT_README


def gen_client_ini_path() -> pathlib.Path:
    """Return path to gen3-client ini file. See https://bit.ly/3NbKGi4"""
    return pathlib.Path(pathlib.Path.home() / ".gen3" / "gen3_client_config.ini")


# TODO - unused, deprecate?
def gen3_client_profile(endpoint: str, path: str = gen_client_ini_path().absolute(), expiration_threshold_days: int = 10) -> str:
    """Read gen3-client ini file, return profile name or none if endpoint not found."""
    gen3_util_ini = read_ini(path)
    matching_sections = []
    for section in gen3_util_ini.sections():
        if gen3_util_ini[section]['api_endpoint'] == endpoint:
            matching_sections.append(section)
    assert len(matching_sections) <= 1, f"Found multiple profiles for {endpoint}: {matching_sections}"
    return matching_sections[0]


def gen3_client_profiles(path: str = gen_client_ini_path().absolute()) -> list[str]:
    """Read gen3-client ini file, return list of profiles."""
    gen3_util_ini = read_ini(path)
    return gen3_util_ini.sections()


def read_ini(path: str):
    """Read ini file."""
    import configparser
    import pathlib

    path = pathlib.Path(path)
    assert path.is_file(), f"{path} is not a file"
    _ = configparser.ConfigParser()
    _.read(path)
    return _


def key_expired_msg(api_key, expiration_threshold_days, key_name):
    """Confirm that api_key is not expired."""
    key = jwt.decode(api_key, options={"verify_signature": False})
    now = datetime.now(tz=timezone.utc).timestamp()
    msg = 'OK, key is valid'
    exp_str = datetime.fromtimestamp(key['exp'], tz=timezone.utc).isoformat()
    iat_str = datetime.fromtimestamp(key['iat'], tz=timezone.utc).isoformat()
    now_str = datetime.fromtimestamp(now, tz=timezone.utc).isoformat()
    if key['exp'] < now:
        msg = f"ERROR key for {key_name} expired {exp_str} < {now_str}"
    if key['iat'] > now:
        msg = f"ERROR key for {key_name} not yet valid {iat_str} > {now_str}"
    delta = timedelta(seconds=key['exp'] - now)
    if 0 < delta.days < expiration_threshold_days:
        msg = f"WARNING {key_name}: Key will expire in {delta.days} days, on {exp_str}"
    return msg


def _get_gen3_client_default_profile(path: pathlib.Path = None, gen3_util_ini: ConfigParser = None) -> str:
    """Read gen3-client ini file, return default (only) profile."""
    if gen3_util_ini is None:
        assert path, "path is required"
        gen3_util_ini = read_ini(path)
    if len(gen3_util_ini.sections()) == 1:
        # default to first section if only one section
        profile = gen3_util_ini.sections()[0]
        return profile
    return None


def _get_gen3_client_key(path: pathlib.Path, profile: str = None) -> str:
    """Read gen3-client ini file, return api_key for profile."""

    gen3_util_ini = read_ini(path)

    if profile:
        for section in gen3_util_ini.sections():
            if section == profile:
                return gen3_util_ini[section]['api_key']
    else:
        profile = _get_gen3_client_default_profile(gen3_util_ini=gen3_util_ini)
        if profile:
            return gen3_util_ini[profile]['api_key']
    click.secho(f"no profile '{profile}' found in {path}, specify one of {gen3_util_ini.sections()}, optionally set environmental variable: GEN3_UTIL_PROFILE", fg='yellow')


def ensure_auth(refresh_file: [pathlib.Path, str] = None, validate: bool = False, config: Config = None) -> Gen3Auth:
    """Confirm connection to Gen3 using their conventions.

    Args:
        refresh_file (pathlib.Path): The file containing the downloaded JSON web token.
        validate: check the connection by getting a new token
        config: Config

    """

    try:
        if refresh_file:
            if isinstance(refresh_file, str):
                refresh_file = pathlib.Path(refresh_file)
            auth = Gen3Auth(refresh_file=refresh_file.name)
        elif 'ACCESS_TOKEN' in os.environ:
            auth = Gen3Auth(refresh_file=f"accesstoken:///{os.getenv('ACCESS_TOKEN')}")
        elif gen_client_ini_path().exists():
            profile = config.gen3.profile
            if not profile:
                # in disconnected mode, or not in project dir
                if config.no_config_found:
                    print("INFO: No config file found in current directory or parents.", file=sys.stderr)
                return None
            # https://github.com/uc-cdis/gen3sdk-python/blob/master/gen3/auth.py#L190-L191
            key = _get_gen3_client_key(gen_client_ini_path(), profile=profile)
            msg = key_expired_msg(key, key_name=profile, expiration_threshold_days=10)
            if 'ERROR' in msg:
                raise ValueError(msg.replace('ERROR', ''))  # remove ERROR prefix
            if 'WARNING' in msg:
                print(msg, file=sys.stderr)
            auth = Gen3Auth(refresh_token={
                'api_key': key,
            })
        else:
            auth = Gen3Auth()

        if validate:
            api_key = auth.refresh_access_token()
            assert api_key, "refresh_access_token failed"

    except (requests.exceptions.ConnectionError, AssertionError) as e:
        msg = (f"Could not get access. profile={profile}"
               "See https://bit.ly/3NbKGi4, or, "
               "store the file in ~/.gen3/credentials.json or specify location with env GEN3_API_KEY "
               f"{e}")

        logging.getLogger(__name__).error(msg)
        raise AssertionError(msg)

    return auth


# TODO - unused, deprecate?
def gen3_services(config: Config) -> tuple[Gen3File, Gen3Index, dict, Gen3Auth]:
    """Create Gen3 Services."""
    auth = ensure_auth(config=config)
    assert auth, f"Failed to set auth {config.gen3.profile}"
    file_client = Gen3File(auth_provider=auth)
    index_client = Gen3Index(auth_provider=auth)
    user = auth.curl('/user/user').json()
    return file_client, index_client, user, auth


def search_upwards_for_file(filename):
    """Search in the current directory and all directories above it
    for a file of a particular name.

    Arguments:
    ---------
    filename :: string, the filename to look for.

    Returns
    -------
    pathlib.Path, the location of the first file found or
    None, if none was found
    """
    d = pathlib.Path.cwd()
    root = pathlib.Path(d.root)

    while d != root:
        attempt = d / filename
        if attempt.exists():
            return attempt
        d = d.parent

    return None


def default():
    """Load config from directory or installed package."""

    # in current dir?
    _ = pathlib.Path(PROJECT_DIR) / 'config.yaml'
    if _.exists():
        return Config(**read_yaml(_))

    # look in parents
    parent_dir = search_upwards_for_file(PROJECT_DIR)
    if parent_dir:
        _ = parent_dir / 'config.yaml'
        if _.exists():
            return Config(**read_yaml(_))

    # use default

    # different pkg_resources open for 3.9
    if sys.version_info[:3] <= (3, 9):
        _config = Config(**yaml.safe_load(pkg_resources.open_text(gen3_tracker, 'config.yaml').read()))
    else:
        # https://docs.python.org/3.11/library/importlib.resources.html#importlib.resources.open_text
        _config = Config(**yaml.safe_load(pkg_resources.files(gen3_tracker).joinpath('config.yaml').open().read()))

    # No config file found in directory or parents.
    _config.no_config_found = True
    return _config


def custom(path: [str, pathlib.Path]):
    """Load user specified config."""
    if isinstance(path, str):
        path = pathlib.Path(path)

    return Config(**read_yaml(path))


def init(config: Config, project_id: str) -> Generator[str, None, None]:
    """Create an empty repository, adjust and write config file"""

    assert project_id, "project_id is missing"
    assert project_id.count('-') == 1, f"{project_id} should have a single '-' delimiter."

    existing_dirs = []
    for _ in PROJECT_DIRECTORIES:
        if pathlib.Path(_).exists():
            existing_dirs.append(_)
        pathlib.Path(_).mkdir(exist_ok=True)

    if existing_dirs:
        yield f"Directory already exists {existing_dirs}"
    else:
        yield f"Created project directories {PROJECT_DIRECTORIES}"

    config.gen3.project_id = project_id
    config.work_dir = pathlib.Path(PROJECT_DIR) / 'work'
    config.work_dir.mkdir(parents=True, exist_ok=True)

    work_dir_git_ignore = config.work_dir / ".gitignore"
    if not pathlib.Path(work_dir_git_ignore).exists():
        with open(work_dir_git_ignore, 'w') as f:
            f.write("*\n!README.md\n!.gitignore\n")

    # meta_dir_git_ignore = "META/.gitignore"
    # if not pathlib.Path(meta_dir_git_ignore).exists():
    #     with open(meta_dir_git_ignore, 'w') as f:
    #         f.write("*\n!README.md")

    for readme in [PROJECT_DIR + '/README.md', "META/README.md"]:
        if not pathlib.Path(readme).exists():
            path = pathlib.Path(readme)
            with open(path, 'w') as f:
                f.write(PROJECT_README)

    config_file = pathlib.Path(PROJECT_DIR) / 'config.yaml'
    if not config_file.exists():
        with open(config_file, 'w') as f:
            yaml.dump(config.model_dump(), f)
        yield f"Created project configuration file={config_file} project_id={config.gen3.project_id} profile={config.gen3.profile}"
    else:
        updated_config = yaml.load(open(config_file), Loader=yaml.SafeLoader)
        updated_config = Config(**updated_config)
        updated_config.gen3.project_id = project_id
        updated_config.gen3.profile = config.gen3.profile
        with open(config_file, 'w') as f:
            yaml.dump(updated_config.model_dump(), f)
        yield f"Updated project configuration file={config_file} project_id={config.gen3.project_id} profile={config.gen3.profile}"



from gen3_tracker.meta import aggregate
from pyvis.network import Network


def _container():
    """Create a pyvis container."""
    return Network(notebook=True, cdn_resources='in_line')  # filter_menu=True, select_menu=True


def _load(net: Network, aggregation: dict) -> Network:
    """Load the aggregation into the visualization network."""
    # add vertices
    for resource_type, _ in aggregation.items():
        assert 'count' in _, _
        net.add_node(resource_type, label=f"{resource_type}/{_['count']}")
    # add edges
    for resource_type, _ in aggregation.items():
        for ref in _.get('references', {}):
            count = _['references'][ref]['count']
            net.add_edge(resource_type, ref, title=count, value=count)
    return net


def create_network_graph(directory_path: str, output_path: str):
    """Render metadata as a network graph into output_path.

    \b
    directory_path: The directory path to the metadata.
    output_path: The path to save the network graph.
    """
    aggregation = aggregate(directory_path)
    # Load it into a pyvis
    net = _load(_container(), aggregation)
    net.save_graph(str(output_path))
    net.show_buttons(filter_=['physics'])



import asyncio
import json
import logging
import pathlib
from datetime import datetime
import time
from urllib.parse import urlparse
from zipfile import ZipFile

from gen3.auth import Gen3Auth
from gen3.jobs import Gen3Jobs
import pytz

from gen3_tracker import Config
from gen3_tracker.common import Push, Commit
from gen3_tracker.gen3.indexd import write_indexd
from gen3_tracker.git import (
    calculate_hash,
    DVC,
    run_command,
    DVCMeta,
    DVCItem,
    modified_date,
)


def _validate_parameters(from_: str) -> pathlib.Path:

    assert (
        len(urlparse(from_).scheme) == 0
    ), f"{from_} appears to be an url. url to url cp not supported"

    return from_


def cp(
    config: Config,
    from_: str,
    project_id: str,
    ignore_state: bool,
    auth=None,
    user=None,
    object_name=None,
    bucket_name=None,
    metadata: dict = {},
):
    """Copy meta to bucket, used by etl_pod job"""
    from_ = _validate_parameters(str(from_))
    if not isinstance(from_, pathlib.Path):
        from_ = pathlib.Path(from_)

    assert auth, "auth is required"

    metadata = dict(
        {"submitter": None, "metadata_version": "0.0.1", "is_metadata": True} | metadata
    )
    if not metadata["submitter"]:
        if not user:
            user = auth.curl("/user/user").json()
        metadata["submitter"] = user["name"]

    program, project = project_id.split("-")

    assert bucket_name, f"could not find bucket for {program}"

    temp_dir = config.work_dir

    temp_dir = pathlib.Path(temp_dir)

    if not object_name:
        now = datetime.now().strftime("%Y%m%d-%H%M%S")
        object_name = f"_{project_id}-{now}_meta.zip"

    zipfile_path = temp_dir / object_name
    with ZipFile(zipfile_path, "w") as zip_object:
        for _ in from_.glob("*.ndjson"):
            zip_object.write(_)

    stat = zipfile_path.stat()
    md5_sum = calculate_hash("md5", zipfile_path)
    my_dvc = DVC(
        meta=DVCMeta(),
        outs=[
            DVCItem(
                path=object_name,
                md5=md5_sum,
                hash="md5",
                modified=modified_date(zipfile_path),
                size=stat.st_size,
            )
        ],
    )

    metadata = write_indexd(
        auth=auth,
        project_id=project_id,
        bucket_name=bucket_name,
        overwrite=False,
        restricted_project_id=None,
        dvc=my_dvc,
    )

    # document = file_client.upload_file_to_guid(guid=id_, file_name=object_name, bucket=bucket_name)
    # print(document, file=sys.stderr)

    run_command(
        f"gen3-client upload-single --bucket {bucket_name} --guid {my_dvc.object_id} --file {zipfile_path} --profile {config.gen3.profile}",
        no_capture=False,
    )

    return {
        "msg": f"Uploaded {zipfile_path} to {bucket_name}",
        "object_id": my_dvc.object_id,
        "object_name": object_name,
    }


def publish_commits(
    config: Config, wait: bool, auth: Gen3Auth, bucket_name: str, spinner=None
) -> dict:
    """Publish commits to the portal."""

    # TODO legacy fhir-import-export job: copies meta to bucket and triggers job,
    #  meta information is already in git REPO,
    #  we should consider changing the fhir_import_export job to use the git REPO

    user = auth.curl("/user/user").json()

    # copy meta to bucket
    upload_result = cp(
        config=config,
        from_="META",
        project_id=config.gen3.project_id,
        ignore_state=True,
        auth=auth,
        user=user,
        bucket_name=bucket_name,
    )

    object_id = upload_result["object_id"]

    push = Push(config=config)
    jobs_client = Gen3Jobs(auth_provider=auth)

    # create "legacy" commit object, read by fhir-import-export job
    push.commits.append(
        Commit(
            object_id=object_id,
            message="From g3t-git",
            meta_path=upload_result["object_name"],
            commit_id=object_id,
        )
    )
    args = {
        "push": push.model_dump(),
        "project_id": config.gen3.project_id,
        "method": "put",
    }

    # capture logging from gen3.jobs
    from cdislogging import get_logger  # noqa

    cdis_logging = get_logger("__name__")
    cdis_logging.setLevel(logging.WARN)

    if wait:
        # async_run_job_and_wait monkeypatched below
        _ = asyncio.run(
            jobs_client.async_run_job_and_wait(
                job_name="fhir_import_export", job_input=args, spinner=spinner
            )
        )
    else:
        _ = jobs_client.create_job("fhir_import_export", args)

    if not isinstance(_, dict):
        _ = {"output": _}
    if isinstance(_["output"], str):
        try:
            _["output"] = json.loads(_["output"])
        except json.JSONDecodeError:
            pass
    return _


# monkey patch for gen3.jobs.Gen3Jobs.async_run_job_and_wait
# make it less noisy and sleep less (max of 30 seconds)
async def async_run_job_and_wait(
    self, job_name, job_input, spinner=None, _ssl=None, **kwargs
):
    """
    Asynchronous function to create a job, wait for output, and return. Will
    sleep in a linear delay until the job is done, starting with 1 second.

    Args:
        _ssl (None, optional): whether or not to use ssl
        job_name (str): name for the job, can use globals in this file
        job_input (Dict): dictionary of input for the job

    Returns:
        Dict: Response from the endpoint
    """
    job_create_response = await self.async_create_job(job_name, job_input)
    status = {"status": "Running", "name": job_name}
    initial_sleep_time = 3
    sleep_time = initial_sleep_time
    max_sleep_time = 30
    while status.get("status") == "Running":
        if spinner:
            spinner.text = f"{status.get('name')} waiting {int(sleep_time)}"
        else:
            logging.debug(f"job still running, waiting for {sleep_time} seconds...")
        time.sleep(sleep_time)
        sleep_time *= 1.5
        if sleep_time > max_sleep_time:
            sleep_time = initial_sleep_time
        status = await self.async_get_status(job_create_response.get("uid"))
        if not spinner:
            logging.info(f"{status}")

    if not spinner:
        logging.info("Job is finished!")
    else:
        spinner.text = f"{status.get('name')} {status.get('status')}"

    if status.get("status") != "Completed":
        # write failed output to log file before raising exception
        response = await self.async_get_output(job_create_response.get("uid"))
        with open("logs/publish.log", "a") as f:
            log_msg = {"timestamp": datetime.now(pytz.UTC).isoformat()}
            log_msg.update(response)
            f.write(json.dumps(log_msg, separators=(",", ":")))
            f.write("\n")

        raise Exception(f"Job status not complete: {status.get('status')}")

    response = await self.async_get_output(job_create_response.get("uid"))
    return response


Gen3Jobs.async_run_job_and_wait = async_run_job_and_wait



import urllib
from os import stat
import tempfile
import pathlib
import requests
from gen3.auth import Gen3Auth
from gen3.file import Gen3File
from zipfile import ZipFile
from urllib.parse import urlparse

import gen3_tracker
from gen3_tracker import Config
from gen3_tracker.gen3.buckets import get_program_bucket
from gen3_tracker.gen3.indexd import write_indexd
from gen3_tracker.git import calculate_hash, DVC, DVCMeta, DVCItem, git_archive, modified_date


def _validate_parameters(from_: str) -> pathlib.Path:

    assert len(urlparse(from_).scheme) == 0, f"{from_} appears to be an url. url to url cp not supported"

    return from_


def push_snapshot(config: Config, auth: Gen3Auth, project_id: str = None, from_: str = None, object_name: str = None):
    """Zip the git repo and push it to the server."""
    # create a zip of the git repo and associate it with the project
    # TODO should we query git to get the list of files to zip?

    proj_id = project_id or config.gen3.project_id
    program, _ = proj_id.split('-')

    # provide support for server provided path name

    if object_name and from_:
        from_ = _validate_parameters(str(from_))
        if not isinstance(from_, pathlib.Path):
            from_ = pathlib.Path(from_)

        temp_dir = tempfile.mkdtemp()
        if from_.is_dir():
            temp_dir = pathlib.Path(temp_dir)
            zipfile_path = temp_dir / object_name
            # double slashes create problems in an s3 environment
            zipfile_path = zipfile_path.name.replace('//', '/')
            with ZipFile(zipfile_path, 'w') as zip_object:
                for _ in from_.glob("*.ndjson"):
                    zip_object.write(_)

    else:
        zipfile_path = str(config.work_dir / f'{config.gen3.project_id}.git.zip')
        git_archive(zipfile_path)

    # this version simply adds the file to indexd and uploads it
    md5_sum = calculate_hash('md5', zipfile_path)
    my_dvc = DVC(
        meta=DVCMeta(),
        project_id=proj_id,
        outs=[
            DVCItem(
                path=str(zipfile_path),
                md5=md5_sum,
                hash='md5',
                modified=modified_date(zipfile_path),
                size=stat(zipfile_path).st_size,
            )
        ]
    )

    if not auth:
        auth = gen3_tracker.config.ensure_auth(config=config)

    bucket_name = get_program_bucket(config=config, program=program, auth=auth)
    _ = write_indexd(
        auth=auth,
        project_id=proj_id,
        bucket_name=bucket_name,
        overwrite=True,
        restricted_project_id=None,
        existing_records=[my_dvc.object_id],
        dvc=my_dvc,
    )

    gen3_file = Gen3File(auth_provider=auth)
    response = gen3_file.upload_file_to_guid(
        bucket=bucket_name,
        guid=my_dvc.object_id,
        file_name=zipfile_path,
        expires_in=3600,
    )
    url = response.get('url', None)
    assert url, f"Expected url in {response}"
    # this url needs to be unquoted
    signed_url = urllib.parse.unquote(url)
    with open(zipfile_path, 'rb') as f:
        files = {'file': (str(zipfile_path), f)}
        # this needs to be a PUT
        response = requests.put(signed_url, files=files)
        response.raise_for_status()

    return {"msg": str(response), "object_id": my_dvc.object_id}

    # cmd = f"gen3-client upload-single --bucket {bucket_name} --guid {my_dvc.object_id} --file {zipfile_path} --profile {config.gen3.profile}",
    # print(cmd)
    # run_command(
    #     cmd,
    #     no_capture=False)



import pathlib
import uuid
from datetime import datetime
from pytz import UTC
from typing import Generator

import orjson
from fhir.resources.attachment import Attachment
from fhir.resources.bundle import Bundle, BundleEntry, BundleEntryRequest
from fhir.resources.documentreference import DocumentReference
from fhir.resources.fhirtypes import DocumentReferenceContentType
from fhir.resources.identifier import Identifier
from fhir.resources.observation import Observation
from fhir.resources.operationoutcome import OperationOutcome
from fhir.resources.patient import Patient
from fhir.resources.researchstudy import ResearchStudy
from fhir.resources.researchsubject import ResearchSubject
from fhir.resources.resource import Resource
from fhir.resources.specimen import Specimen
from fhir.resources.task import Task, TaskOutput, TaskInput

from gen3_tracker import ACED_NAMESPACE
from gen3_tracker.common import create_resource_id, EmitterContextManager
from gen3_tracker.git import DVC, run_command, dvc_data


def _get_system(identifier: str, project_id: str):
    """Return system component of simplified identifier"""
    if '#' in identifier:
        return identifier.split('#')[0]
    if '|' in identifier:
        return identifier.split('|')[0]
    # default
    return f"https://aced-idp.org/{project_id}"


def meta_index():
    """Read all the ndjson files in the `META` directory and create a dictionary with the id as the key and the official identifier as the value"""
    meta_dir = pathlib.Path('META')
    id_dict = {}

    for file in meta_dir.glob('*.ndjson'):
        with open(file, 'r') as f:
            for line in f:
                record = orjson.loads(line)
                _id = record.get('id')
                resource_type = record.get('resourceType')
                if resource_type == 'Bundle':
                    break
                official_identifier = next((identifier.get('value') for identifier in record.get('identifier', []) if identifier.get('use') == 'official'), None)
                if not official_identifier and record.get('identifier'):
                    official_identifier = record['identifier'][0]['value']

                if _id and official_identifier:
                    id_dict[f"{resource_type}/{_id}"] = official_identifier

    return id_dict


def get_data_from_meta() -> Generator[int, None, None]:
    """Read all the ndjson files in the `META` directory and return a generator that produces all records"""
    meta_dir = pathlib.Path('META')

    for file in meta_dir.glob('*.ndjson'):
        with open(file, 'r') as f:
            for line in f:
                record = orjson.loads(line)
                resource_type = record.get('resourceType')
                if resource_type == 'Bundle':
                    break

                yield record


def update_document_reference(document_reference: DocumentReference, dvc_data: DVC):
    """Update document reference with index record."""
    assert document_reference.resource_type == 'DocumentReference'
    assert dvc_data.out.object_id == document_reference.id, f"{dvc_data['did']} != {document_reference.id}"
    assert dvc_data.out.modified, f"dvc_data missing modified: {dvc_data}"
    document_reference.docStatus = 'final'
    document_reference.status = 'current'

    document_reference.date = dvc_data.out.modified

    attachment = Attachment()
    if dvc_data.out.realpath:
        source_path = f"file:///{dvc_data.out.realpath}"
    else:
        source_path = dvc_data.out.source_url

    source_path = source_path.replace('////', '///')

    attachment.extension = [
        {
            "url": f"http://aced-idp.org/fhir/StructureDefinition/{dvc_data.out.hash}",
            "valueString": dvc_data.out.hash_value
        },
        {
            "url": "http://aced-idp.org/fhir/StructureDefinition/source_path",
            "valueUrl": source_path
        }
    ]
    attachment.contentType = dvc_data.out.mime

    attachment.url = f"file:///{dvc_data.out.path}"
    # TODO "url": f"file:///{str(file).replace(remove_path_prefix, '', 1)}",  # Uri where the data can be found
    attachment.size = dvc_data.out.size
    attachment.title = pathlib.Path(dvc_data.out.path).name
    attachment.creation = dvc_data.out.modified

    content = DocumentReferenceContentType(attachment=attachment)

    document_reference.content = [content]


def create_id_from_strings(resource_type: str, project_id: str, identifier_string: str) -> str:
    """Create an id from strings."""
    if not identifier_string:
        return None
    return str(uuid.uuid5(ACED_NAMESPACE, f"{project_id}/{resource_type}/{_get_system(identifier_string, project_id)}|{identifier_string}"))


def create_skeleton(dvc: dict, project_id: str, meta_index: set[str] = []) -> list[Resource]:
    """
    Create a skeleton graph for document and ancestors from a set of identifiers.
    """
    dvc = DVC.model_validate(dvc)

    specimen_identifier = dvc.meta.specimen
    patient_identifier = dvc.meta.patient
    task_identifier = dvc.meta.task
    observation_identifier = dvc.meta.observation
    project_id = dvc.project_id or project_id

    assert dvc.out, f"out required {dvc}"
    document_reference_id = dvc.out.set_object_id(project_id=project_id)

    assert project_id, "project_id required"
    assert project_id.count('-') == 1, "project_id must be of the form program-project"
    program, project = project_id.split('-')

    research_study = research_subject = observation = specimen = patient = task = document_reference = None

    # check if we have already created the resources

    research_study_id = create_id_from_strings(resource_type='ResearchStudy', project_id=project_id, identifier_string=project_id)
    specimen_id = create_id_from_strings(resource_type='Specimen', project_id=project_id, identifier_string=specimen_identifier)
    patient_id = create_id_from_strings(resource_type='Patient', project_id=project_id, identifier_string=patient_identifier)
    task_id = create_id_from_strings(resource_type='Task', project_id=project_id, identifier_string=task_identifier)
    observation_id = create_id_from_strings(resource_type='Observation', project_id=project_id, identifier_string=observation_identifier)

    _ = f'ResearchStudy/{research_study_id}'
    if _ in meta_index:
        research_study = meta_index[_]
    _ = f'Specimen/{specimen_id}'
    if _ in meta_index:
        specimen = meta_index[_]

    _ = f'Patient/{patient_id}'
    if _ in meta_index:
        patient = meta_index[_]

    _ = f'Task/{task_id}'
    if _ in meta_index:
        task = meta_index[_]

    _ = f'Observation/{observation_id}'
    if _ in meta_index:
        observation = meta_index[_]

    # create entities

    document_reference = DocumentReference(status='current', content=[{'attachment': {'url': "file://"}}])
    document_reference.id = document_reference_id
    document_reference.identifier = [

        Identifier(value=document_reference_id, system=_get_system(document_reference_id, project_id=project_id),
                   use='official')]
    update_document_reference(document_reference, dvc)

    if not research_study:
        research_study = ResearchStudy(status='active')
        research_study.description = f"Skeleton ResearchStudy for {project_id}"
        research_study.identifier = [
            Identifier(value=project_id, system=_get_system(project_id, project_id=project_id),
                       use='official')]
        research_study.id = create_resource_id(research_study, project_id)

    if not patient and patient_identifier:
        patient = Patient()
        patient.identifier = [Identifier(value=patient_identifier, system=_get_system(patient_identifier, project_id=project_id), use='official')]
        patient.id = create_resource_id(patient, project_id)

        research_subject = ResearchSubject(
            status='active',
            study={'reference': f"ResearchStudy/{research_study_id}"},
            subject={'reference': f"Patient/{patient.id}"}
        )
        research_subject.identifier = [Identifier(value=patient_identifier, system=_get_system(patient_identifier, project_id=project_id), use='official')]
        research_subject.id = create_resource_id(research_subject, project_id)
        patient_id = patient.id

    if not observation and observation_identifier:
        observation = Observation(status='final', code={'text': 'unknown'})
        observation.identifier = [Identifier(value=observation_identifier, system=_get_system(observation_identifier, project_id=project_id), use='official')]
        observation.id = create_resource_id(observation, project_id)

        assert patient, "patient required for observation"
        observation.subject = {'reference': f"Patient/{patient_id}"}

    if not specimen and specimen_identifier:

        # TODO:?
        # print(f'{specimen_identifier} Specimen/{specimen_id} not in {[(k, _) for k, _ in meta_index.items() if k.startswith("Specimen")]}')
        # exit(1)

        specimen = Specimen()
        specimen.identifier = [Identifier(value=specimen_identifier, system=_get_system(specimen_identifier, project_id=project_id), use='official')]
        specimen.id = create_resource_id(specimen, project_id)
        specimen_id = specimen.id

        assert patient, "patient required for specimen"
        specimen.subject = {'reference': f"Patient/{patient_id}"}

    if not task and task_identifier:
        task = Task(intent='unknown', status='completed')
        task.identifier = [Identifier(value=task_identifier, system=_get_system(task_identifier, project_id=project_id),
                                      use='official')]
        task.id = create_resource_id(task, project_id)
        task_id = task.id

    # create relationships

    # assign subject, specimen of observation
    if observation and specimen and not observation.specimen:
        observation.specimen = {'reference': f"Specimen/{specimen_id}"}
    if observation and patient and not observation.subject:
        observation.subject = {'reference': f"Patient/{patient_id}"}

    if task:
        task.input = []
        if specimen:
            task.input.append(
                TaskInput(
                    valueReference={"reference": f"Specimen/{specimen.id}"},
                    type={'text': 'Specimen'}
                )
            )
        if patient:
            task.input.append(
                TaskInput(
                    valueReference={'reference': f"Patient/{patient.id}"},
                    type={'text': 'Patient'}
                )
            )
        task.output = [
            TaskOutput(
                valueReference={'reference': f"DocumentReference/{document_reference.id}"},
                type={'text': 'DocumentReference'})
        ]

    # assign document reference subject
    if observation:
        document_reference.subject = {'reference': f"Observation/{observation_id}"}
    if specimen and not document_reference.subject:
        document_reference.subject = {'reference': f"Specimen/{specimen_id}"}
    if patient and not document_reference.subject:
        document_reference.subject = {'reference': f"Patient/{patient_id}"}
    if not document_reference.subject:
        document_reference.subject = {'reference': f"ResearchStudy/{research_study_id}"}

    return [_ for _ in [research_study, research_subject, patient, observation, specimen, task, document_reference] if _ and not isinstance(_, str)]


def update_meta_files(dry_run=False, project_id=None) -> list[str]:
    """Maintain the META directory."""
    assert project_id, "project_id required"
    manifest_path = pathlib.Path('MANIFEST')
    dvc_files = [_ for _ in manifest_path.rglob('*.dvc')]

    if not dvc_files:
        return []

    before_meta_files = [_ for _ in pathlib.Path('META').glob('*.ndjson')]
    before_meta_index = set(list(meta_index().keys()))
    emitted_already = []

    with EmitterContextManager('META') as emitter:
        for _ in dvc_data(dvc_files):
            resources = create_skeleton(_, project_id, meta_index())
            for resource in resources:
                key = f"{resource.resource_type}/{resource.id}"
                if key not in emitted_already:
                    emitter.emit(resource.resource_type).write(
                        resource.json(option=orjson.OPT_APPEND_NEWLINE)
                    )
                    emitted_already.append(key)

    after_meta_index = set(list(meta_index().keys()))
    orphaned_meta_index = before_meta_index - after_meta_index

    if orphaned_meta_index:
        # create a bundle to tell server about deletes
        now = datetime.now(UTC).strftime("%Y-%m-%dT%H:%M:%S.%fZ")
        bundle = Bundle(type='transaction', timestamp=now)

        bundle.identifier = Identifier(value=project_id, system="https://aced-idp.org/project_id", use='official')
        bundle.id = create_resource_id(bundle, project_id)

        bundle.entry = []
        outcome = OperationOutcome(issue=[{'severity': 'warning', 'code': 'processing', 'diagnostics': 'Meta data items no longer in study.'}])
        bundle.issues = outcome

        for _ in orphaned_meta_index:
            bundle_entry = BundleEntry()
            bundle_entry.request = BundleEntryRequest(url=_, method='DELETE')
            bundle.entry.append(bundle_entry)

        with EmitterContextManager('META') as emitter:
            emitter.emit(bundle.resource_type, file_mode='a').write(
                bundle.json(option=orjson.OPT_APPEND_NEWLINE)
            )

    after_meta_files = [_ for _ in pathlib.Path('META').glob('*.ndjson')]
    new_meta_files = [str(_) for _ in after_meta_files if _ not in before_meta_files]

    if new_meta_files:
        run_command(f'git add {" ".join(new_meta_files)}', dry_run=dry_run, no_capture=True)

    return after_meta_files



import glob
import json
import logging
import multiprocessing
import os
import pathlib
import uuid
from collections import defaultdict

import orjson
import re
import shutil
import subprocess
import requests
import sys
import time
import zipfile
from datetime import datetime
from pytz import UTC


import click
import pytz
import yaml
from fhir.resources.documentreference import DocumentReference
from fhir.resources.identifier import Identifier
from fhir.resources.bundle import Bundle, BundleEntry, BundleEntryRequest

from gen3.auth import Gen3AuthError
from gen3.file import Gen3File
from gen3.index import Gen3Index
from halo import Halo
from tqdm import tqdm

import gen3_tracker
from gen3_tracker import Config
from gen3_tracker.common import (
    CLIOutput,
    INFO_COLOR,
    ERROR_COLOR,
    is_url,
    filter_dicts,
    SUCCESS_COLOR,
    read_ndjson_file,
)
from gen3_tracker.config import init as config_init, ensure_auth
from gen3_tracker.gen3.buckets import get_buckets
from gen3_tracker.git import (
    git_files,
    to_indexd,
    to_remote,
    dvc_data,
    data_file_changes,
    modified_date,
    git_status,
    DVC,
    MISSING_G3T_MESSAGE,
)
from gen3_tracker.git import run_command, MISSING_GIT_MESSAGE, git_repository_exists
from gen3_tracker.git.adder import url_path, write_dvc_file
from gen3_tracker.git.cloner import ls
from gen3_tracker.git.initializer import initialize_project_server_side
from gen3_tracker.git.snapshotter import push_snapshot
from gen3_tracker.meta.skeleton import meta_index, get_data_from_meta
from gen3_tracker.common import _default_json_serializer

# logging.basicConfig(level=logging.INFO)
_logger = logging.getLogger(__package__)


# @click.command(cls=NaturalOrderGroup)
# @click.option('--debug', is_flag=True, envvar='G3T_DEBUG', help='Enable debug mode. G3T_DEBUG environment variable can also be used.')
# @click.option('--dry-run', is_flag=True, envvar='G3T_DRYRUN', help='Print the commands that would be executed, but do not execute them. G3T_DRYRUN environment variable can also be used.')
# @click.version_option()
# @click.pass_obj
# def cli(ctx: click.Context, debug: bool, dry_run: bool):
#     """[Experimental] A CLI for adding version control to Gen3 projects."""
#
#     config.ensure_object(dict)
#     config.debug = debug
#     config.debug = dry_run
#     if debug:
#         _logger.setLevel(logging.DEBUG)


def _check_parameters(config, project_id):
    """Common parameter checks."""
    if not project_id:
        raise AssertionError("project_id is required")
    if not project_id.count("-") == 1:
        raise AssertionError(
            f"project_id must be of the form program-project {project_id}"
        )
    if not config.gen3.profile:
        click.secho(
            "No profile set. Continuing in disconnected mode. Use `set profile <profile>`",
            fg="yellow",
        )


@click.group(cls=gen3_tracker.NaturalOrderGroup)
def cli():
    """git-like version control for Gen3 projects."""
    pass


@cli.command(context_settings=dict(ignore_unknown_options=True))
# @click.option('--force', '-f', is_flag=True, help='Force the init.')
@click.argument(
    "project_id",
    default=None,
    required=False,
    envvar=f"{gen3_tracker.ENV_VARIABLE_PREFIX}PROJECT_ID",
)
@click.option(
    "--approve",
    "-a",
    help="Approve the addition (privileged)",
    is_flag=True,
    default=False,
    show_default=True,
)
@click.option(
    "--no-server",
    help="Skip server setup (testing)",
    is_flag=True,
    default=False,
    show_default=True,
    hidden=True,
)
@click.option(
    "--debug",
    is_flag=True,
    envvar="G3T_DEBUG",
    help="Enable debug mode. G3T_DEBUG environment variable can also be used.",
)
@click.pass_obj
def init(config: Config, project_id: str, approve: bool, no_server: bool, debug: bool):
    """Initialize a new repository."""
    try:
        # uncomment if we want to check for a git remote
        # try:
        #     assert not git_remote_exists(config.debug), 'A git remote exists. PHI and large data files should not be stored in a public repository. Use the --force option to continue.'
        # except Exception as e:
        #     if not force:
        #         click.secho(str(e), fg=ERROR_COLOR, file=sys.stderr)
        #         if config.debug:
        #             raise

        # click.secho(INIT_MESSAGE, fg=INFO_COLOR, file=sys.stderr)

        _check_parameters(config, project_id)

        logs = []
        # create directories
        for _ in config_init(config, project_id):
            logs.append(_)

        ensure_git_repo(config)

        if not no_server:
            init_logs, approval_needed = initialize_project_server_side(
                config, project_id
            )
            logs.extend(init_logs)

            if approve and approval_needed:
                run_command(
                    "g3t collaborator approve --all",
                    dry_run=config.dry_run,
                    no_capture=True,
                )
            elif approval_needed and not approve:
                click.secho(
                    "Approval needed. to approve the project, a privileged user must run `g3t collaborator approve --all`",
                    fg=INFO_COLOR,
                    file=sys.stderr,
                )
            else:
                click.secho(
                    f"Approval not needed. Project {project_id} has approved read/write",
                    fg=INFO_COLOR,
                    file=sys.stderr,
                )

        if config.debug:
            for _ in logs:
                click.secho(_, fg=INFO_COLOR, file=sys.stderr)

    except Exception as e:
        click.secho(str(e), fg=ERROR_COLOR, file=sys.stderr)
        if config.debug or debug:
            raise


def ensure_git_repo(config):
    # ensure a git repo
    if pathlib.Path(".git").exists():
        return

    if not pathlib.Path(".git").exists():
        command = "git init"
        run_command(command, dry_run=config.dry_run, no_capture=True)
    else:
        click.secho("Git repository already exists.", fg=INFO_COLOR, file=sys.stderr)
    pathlib.Path("MANIFEST").mkdir(exist_ok=True)
    pathlib.Path("META").mkdir(exist_ok=True)
    pathlib.Path("LOGS").mkdir(exist_ok=True)
    with open(".gitignore", "w") as f:
        f.write("/LOGS/\n")
        f.write(".g3t/state/\n")  # legacy
    with open("META/README.md", "w") as f:
        f.write(
            "This directory contains metadata files for the data files in the MANIFEST directory.\n"
        )
    with open("MANIFEST/README.md", "w") as f:
        f.write("This directory contains dvc files that reference the data files.\n")
    run_command(
        "git add MANIFEST META .gitignore .g3t", dry_run=config.dry_run, no_capture=True
    )
    run_command(
        'git commit -m "initialized" MANIFEST META .gitignore .g3t',
        dry_run=config.dry_run,
        no_capture=True,
    )


# Note: The commented code below is an example of how to use context settings to allow extra arguments.
# context_settings=dict(
#     ignore_unknown_options=True,
#     allow_extra_args=True,
# )


@cli.command(context_settings=dict(ignore_unknown_options=True, allow_extra_args=True))
@click.argument("target")
@click.option("--no-git-add", default=False, is_flag=True, hidden=True)
@click.pass_context
def add(ctx, target, no_git_add: bool):
    """
    Update references to data files to the repository.

    \b
    TARGET is a "data file", either files or urls.
    We commit their proxies - MANIFEST/<TARGET>.dvc files.
    \b
    If the TARGET is a file in the local file system:
        - We will automatically calculate the hash, size, modified and mime type of the file
        - You can specify those values with the --<hash>, --size, --modified and --mime options
        - As a convenience, you can use wildcards to add multiple files at once.
          If wildcards are used, the hash, size, modified and mime type parameters are ignored.
    \b
    If the TARGET is a url:
        - You must specify the hash, size, modified and mime type
        - Wildcards are not supported
    \b
    --<hash> <value>: Valid options are: ['md5', 'sha1', 'sha256', 'sha512', 'crc', 'etag']
                      Value must conform to the hash type.
    --modified: A variety of date formats are supported, see https://tinyurl.com/ysad3rj7
    --mime: If not specified, it will be inferred from the file extension.
    --no-bucket: If specified, the file will not be uploaded to the bucket, and user will access via scp or symlink.
    --no-git-add: If specified, the file will not be automatically added to git. Avoids locking, useful for parallel adds .
    \b
    Identifiers:
    In order to link a file with associated Patient, Specimen, Observation or Task, you can use one of the following identifiers:
    --patient <value>: The patient identifier
    --specimen <value>: The specimen identifier, requires patient
    --observation <value>: The observation identifier, requires patient, optionally specimen
    --task <value>: The task identifier, requires either patient or specimen
    The <value> is a user defined string that will be used to link the data file with the associated resource.  Do not use PHI in the value.
    See `g3t meta` for more
    """
    from gen3_tracker.git.adder import add_file, add_url

    config: Config = ctx.obj
    try:
        # needs to be in project root
        assert git_repository_exists(config.debug), MISSING_GIT_MESSAGE
        assert not config.no_config_found, MISSING_G3T_MESSAGE

        # needs to have a target
        assert target, "No targets specified."

        # Expand wildcard paths
        if is_url(target) and not target.startswith("file://"):
            all_changed_files, updates = add_url(ctx, target)
        else:
            all_changed_files, updates = add_file(ctx, target)

        #
        # if it is an update, we do not need to add the file to git
        #
        adds = [str(_) for _ in all_changed_files if _ not in updates]
        if adds and not no_git_add:
            adds.append(".gitignore")
            run_command(
                f'git add {" ".join([str(_) for _ in adds])}',
                dry_run=config.dry_run,
                no_capture=True,
            )

    except Exception as e:
        click.secho(str(e), fg=ERROR_COLOR, file=sys.stderr)
        if config.debug:
            raise


@cli.command(context_settings=dict(ignore_unknown_options=True, allow_extra_args=True))
@click.argument("targets", nargs=-1)
@click.option("--message", "-m", help="The commit message.")
@click.option(
    "--all",
    "-a",
    is_flag=True,
    default=False,
    help="Automatically stage files that have been modified and deleted.",
)
@click.pass_context
def commit(ctx, targets, message, all):
    """Commit the changes
    \b
    TARGETS: list of files (default: all)
    """
    config: Config = ctx.obj
    command = [
        "git",
        "commit",
        "-m",
        f'"{message}"',
    ] + list(targets)

    if all:
        command.append("-a")

    try:
        run_command(" ".join(command), dry_run=config.dry_run, no_capture=True)
    except Exception as e:
        click.secho(str(e), fg=ERROR_COLOR, file=sys.stderr)
        if config.debug:
            raise
        exit(1)


@cli.command()
@click.pass_obj
def status(config):
    """Show changed files."""
    soft_error = False
    try:
        with Halo(text="Scanning", spinner="line", placement="right", color="white"):
            manifest_path = pathlib.Path("MANIFEST")
            changes = data_file_changes(manifest_path)
            # Get a list of all files in the MANIFEST directory and its subdirectories
            files = glob.glob("MANIFEST/**/*.dvc", recursive=True)
            # Filter out directories, keep only files
            files = [f for f in files if os.path.isfile(f)]
        if not files:
            click.secho("No files have been added.", fg=INFO_COLOR, file=sys.stderr)
        else:
            # Find the most recently changed file
            latest_file = max(files, key=os.path.getmtime)

            document_reference_mtime = 0

            if pathlib.Path("META/DocumentReference.ndjson").exists():
                # Get the modification time
                document_reference_mtime = os.path.getmtime(
                    "META/DocumentReference.ndjson"
                )

            latest_file_mtime = os.path.getmtime(latest_file)
            if document_reference_mtime < latest_file_mtime:
                document_reference_mtime = datetime.fromtimestamp(
                    document_reference_mtime
                ).isoformat()
                latest_file_mtime = datetime.fromtimestamp(
                    latest_file_mtime
                ).isoformat()
                click.secho(
                    f"WARNING: DocumentReference.ndjson is out of date {document_reference_mtime}. The most recently changed file is {latest_file} {latest_file_mtime}.  Please check DocumentReferences.ndjson",
                    fg=INFO_COLOR,
                    file=sys.stderr,
                )
                soft_error = True

            if changes:
                click.secho(
                    f"# There are {len(changes)} data files that you need to update via `g3t add`:",
                    fg=INFO_COLOR,
                    file=sys.stderr,
                )
                cwd = pathlib.Path.cwd()
                for _ in changes:
                    data_path = str(_.data_path).replace(str(cwd) + "/", "")
                    click.secho(
                        f"  g3t add {data_path} # changed: {modified_date(_.data_path)},  last added: {modified_date(_.dvc_path)}",
                        fg=INFO_COLOR,
                        file=sys.stderr,
                    )
                    soft_error = True
            else:
                click.secho("No data file changes.", fg=INFO_COLOR, file=sys.stderr)

        _ = run_command("git status")
        print(_.stdout)
        if soft_error:
            exit(1)
    except Exception as e:
        click.secho(str(e), fg=ERROR_COLOR, file=sys.stderr)
        if config.debug:
            raise


@cli.command()
@click.option(
    "--step",
    type=click.Choice(["index", "upload", "publish", "all", "fhir"]),
    default="all",
    show_default=True,
    help="The step to run ",
)
@click.option(
    "--transfer-method",
    type=click.Choice(gen3_tracker.FILE_TRANSFER_METHODS.keys()),
    default="gen3",
    show_default=True,
    help="The upload method.",
)
@click.option(
    "--overwrite", is_flag=True, help="(index): Overwrite previously submitted files."
)
@click.option(
    "--wait",
    default=True,
    is_flag=True,
    show_default=True,
    help="(publish): Wait for metadata completion.",
)
@click.option(
    "--dry-run",
    show_default=True,
    default=False,
    is_flag=True,
    help="Print the commands that would be executed, but do not execute them.",
)
@click.option(
    "--re-run",
    show_default=True,
    default=False,
    is_flag=True,
    help="Re-run the last publish step",
)
@click.option(
    "--fhir-server",
    show_default=True,
    default=False,
    is_flag=True,
    help="Push data in META directory to FHIR Server. Whatever FHIR data that exists in META dir will be upserted into the fhir server",
)
@click.option("--debug", is_flag=True)
@click.option("--skip_validate", is_flag=True, help="Skip validation of the metadata")
@click.pass_context
def push(
    ctx,
    step: str,
    transfer_method: str,
    overwrite: bool,
    re_run: bool,
    wait: bool,
    dry_run: bool,
    fhir_server: bool,
    debug: bool,
    skip_validate: bool,
):
    """Push changes to the remote repository.
    \b
    steps:
        index - index the files
        push - push the files to the remote
        publish - publish the files to the portal
        all - all of the above.
    transfer-method: specify the remote storage type:
        gen3 - gen3-client to/from local
        no-bucket - indexd only symlink to/from local
        s3 - (admin) s3 to/from local
        s3-map - (admin) s3 index only external s3
    """
    from gen3_tracker.gen3.jobs import publish_commits
    from gen3_tracker.gen3.buckets import get_program_bucket

    config = ctx.obj

    try:

        if re_run:
            # step = 'publish'
            raise NotImplementedError("Re-run not implemented")

        try:
            with Halo(
                text="Checking", spinner="line", placement="right", color="white"
            ):
                run_command("g3t status")
                if not skip_validate:
                    run_command("g3t meta validate", no_capture=True)

        except Exception as e:
            click.secho(
                "Please correct issues before pushing.", fg=ERROR_COLOR, file=sys.stderr
            )
            click.secho(str(e), fg=ERROR_COLOR, file=sys.stderr)
            if config.debug:
                raise
            exit(1)

        with Halo(text="Scanning", spinner="line", placement="right", color="white"):

            # check git status
            branch, uncommitted = git_status()
            assert (
                not uncommitted
            ), "Uncommitted changes found.  Please commit or stash them first."

            # check dvc vs external files
            changes = data_file_changes(pathlib.Path("MANIFEST"))
            assert (
                not changes
            ), f"# There are {len(changes)} data files that you need to update.  See `g3t status`"

            # initialize dvc objects with this project_id
            # dvc objects are retrieved only if committed to git
            committed_files, dvc_objects = manifest(config.gen3.project_id)

            # initialize gen3 client
            auth = gen3_tracker.config.ensure_auth(config=config)
            bucket_name = get_program_bucket(config=config, auth=auth)

            # check for new files
            records = ls(
                config, metadata={"project_id": config.gen3.project_id}, auth=auth
            )["records"]
            dids = {_["did"]: _["updated_date"] for _ in records}
            new_dvc_objects = [_ for _ in dvc_objects if _.object_id not in dids]
            updated_dvc_objects = [
                _
                for _ in dvc_objects
                if _.object_id in dids and _.out.modified > dids[_.object_id]
            ]
            if step not in ["publish", "fhir"]:
                if not overwrite:
                    dvc_objects = new_dvc_objects + updated_dvc_objects
                    assert (
                        dvc_objects
                    ), "No new files to index.  Use --overwrite to force"

        click.secho(
            f"Scanned new: {len(new_dvc_objects)}, updated: {len(updated_dvc_objects)} files",
            fg=INFO_COLOR,
            file=sys.stderr,
        )
        if updated_dvc_objects:
            click.secho(
                f"Found {len(updated_dvc_objects)} updated files. overwriting",
                fg=INFO_COLOR,
                file=sys.stderr,
            )
            overwrite = True

        if step in ["index", "all"]:
            # send to index

            if dry_run:
                click.secho(
                    "Dry run: not indexing files", fg=INFO_COLOR, file=sys.stderr
                )
                yaml.dump(
                    {
                        "new": [_.model_dump() for _ in new_dvc_objects],
                        "updated": [_.model_dump() for _ in updated_dvc_objects],
                    },
                    sys.stdout,
                )
                return

            for _ in tqdm(
                to_indexd(
                    dvc_objects=dvc_objects,
                    auth=auth,
                    project_id=config.gen3.project_id,
                    bucket_name=bucket_name,
                    overwrite=overwrite,
                    restricted_project_id=None,
                ),
                desc="Indexing",
                unit="file",
                leave=False,
                total=len(committed_files),
            ):
                pass
            click.secho(
                f"Indexed {len(committed_files)} files.", fg=INFO_COLOR, file=sys.stderr
            )

        if step in ["upload", "all"]:
            click.secho(
                f"Checking {len(dvc_objects)} files for upload via {transfer_method}",
                fg=INFO_COLOR,
                file=sys.stderr,
            )
            to_remote(
                upload_method=transfer_method,
                dvc_objects=dvc_objects,
                bucket_name=bucket_name,
                profile=config.gen3.profile,
                dry_run=config.dry_run,
                work_dir=config.work_dir,
            )

        if fhir_server or step in ["fhir"]:
            """Either there exists a Bundle.ndjson file in META signifying a revision to the data, or there is no bundle.json,
            signifying that the data in the META directory should be upserted into gen34
            """
            meta_dir = pathlib.Path("META")
            bundle_file = meta_dir / "Bundle.ndjson"
            if os.path.isfile(bundle_file):
                with Halo(
                    text="Sending to FHIR Server",
                    spinner="line",
                    placement="right",
                    color="white",
                ):
                    with open(bundle_file, "r") as file:
                        json_string = file.read()
                    bundle_data = orjson.loads(json_string)
                    headers = {"Authorization": f"{auth._access_token}"}
                    result = requests.delete(
                        url=f"{auth.endpoint}/Bundle",
                        data=orjson.dumps(
                            bundle_data,
                            default=_default_json_serializer,
                            option=orjson.OPT_APPEND_NEWLINE,
                        ).decode(),
                        headers=headers,
                    )

                with open("logs/publish.log", "a") as f:
                    log_msg = {
                        "timestamp": datetime.now(pytz.UTC).isoformat(),
                        "result": f"{result}",
                    }
                    click.secho(
                        "Published project. See logs/publish.log",
                        fg=SUCCESS_COLOR,
                        file=sys.stderr,
                    )
                    f.write(json.dumps(log_msg, separators=(",", ":")))
                    f.write("\n")
                return

            project_id = config.gen3.project_id
            now = datetime.now(UTC).strftime("%Y-%m-%dT%H:%M:%S.%fZ")
            bundle = Bundle(type="transaction", timestamp=now)
            bundle.identifier = Identifier(
                value=project_id, system="https://aced-idp.org/project_id"
            )
            from gen3_tracker import ACED_NAMESPACE

            bundle.id = str(uuid.uuid5(ACED_NAMESPACE, f"Bundle/{project_id}/{now}"))
            bundle.entry = []

            for _ in get_data_from_meta():
                bundle_entry = BundleEntry()
                # See https://build.fhir.org/bundle-definitions.html#Bundle.entry.request.url
                bundle_entry.request = BundleEntryRequest(
                    url=f"{_['resourceType']}/{_['id']}", method="PUT"
                )
                bundle_entry.resource = _
                bundle.entry.append(bundle_entry)

            headers = {"Authorization": f"{auth._access_token}"}
            bundle_dict = bundle.dict()
            with Halo(
                text="Sending to FHIR Server",
                spinner="line",
                placement="right",
                color="white",
            ):
                result = requests.put(
                    url=f"{auth.endpoint}/Bundle",
                    data=orjson.dumps(
                        bundle_dict,
                        default=_default_json_serializer,
                        option=orjson.OPT_APPEND_NEWLINE,
                    ).decode(),
                    headers=headers,
                )

            with open("logs/publish.log", "a") as f:
                log_msg = {
                    "timestamp": datetime.now(pytz.UTC).isoformat(),
                    "result": f"{result}",
                }
                click.secho(
                    "Published project. See logs/publish.log",
                    fg=SUCCESS_COLOR,
                    file=sys.stderr,
                )
                f.write(json.dumps(log_msg, separators=(",", ":")))
                f.write("\n")
            return

        if step in ["publish", "all"] and not fhir_server:
            log_path = "logs/publish.log"

            with Halo(
                text="Uploading snapshot",
                spinner="line",
                placement="right",
                color="white",
            ):
                # push the snapshot of the `.git` sub-directory in the current directory
                push_snapshot(config, auth=auth)

            if transfer_method == "gen3":
                try:
                    # legacy, "old" fhir_import_export use publish_commits to publish the META
                    with Halo(
                        text="Publishing",
                        spinner="line",
                        placement="right",
                        color="white",
                    ) as spinner:
                        _ = publish_commits(
                            config,
                            wait=wait,
                            auth=auth,
                            bucket_name=bucket_name,
                            spinner=spinner,
                        )
                except Exception as e:
                    click.secho(
                        f"Unable to publish project. See {log_path} for more info",
                        fg=ERROR_COLOR,
                        file=sys.stderr,
                    )
                    raise e

                # print success message and save logs
                with open(log_path, "a") as f:
                    log_msg = {"timestamp": datetime.now(pytz.UTC).isoformat()}
                    log_msg.update(_)
                    f.write(json.dumps(log_msg, separators=(",", ":")))
                    f.write("\n")
                click.secho(
                    f"Published project. Logs found at {log_path}",
                    fg=SUCCESS_COLOR,
                    file=sys.stderr,
                )
            else:
                click.secho(
                    f"Auto-publishing not supported for {transfer_method}. Please use --step publish after uploading",
                    fg=ERROR_COLOR,
                    file=sys.stderr,
                )

    except Exception as e:
        click.secho(str(e), fg=ERROR_COLOR, file=sys.stderr)
        if config.debug or debug:
            raise
        exit(1)


def manifest(project_id) -> tuple[list[str], list[DVC]]:
    """Get the committed files and their dvc objects. Initialize dvc objects with this project_id"""
    committed_files = [_ for _ in git_files() if _.endswith(".dvc")]
    dvc_objects = [_ for _ in dvc_data(committed_files)]
    for _ in dvc_objects:
        _.project_id = project_id
    return committed_files, dvc_objects


@cli.command()
@click.option(
    "--remote",
    type=click.Choice(["gen3", "s3", "ln", "scp"]),
    default="gen3",
    show_default=True,
    help="Specify the remote storage type. gen3:download, s3:s3 cp, ln: symbolic link, scp: scp copy",
)
@click.option(
    "--worker_count",
    "-w",
    default=(multiprocessing.cpu_count() - 1),
    show_default=True,
    type=int,
    help="Number of workers to use.",
)
@click.option(
    "--data-only",
    help="Ignore git snapshot",
    is_flag=True,
    default=False,
    show_default=True,
)
@click.pass_obj
def pull(config: Config, remote: str, worker_count: int, data_only: bool):
    """Fetch from and integrate with a remote repository."""
    try:

        with Halo(text="Authorizing", spinner="line", placement="right", color="white"):
            auth = gen3_tracker.config.ensure_auth(config=config)

        if not data_only:
            with Halo(
                text="Pulling git snapshot",
                spinner="line",
                placement="right",
                color="white",
            ):
                if not auth:
                    auth = gen3_tracker.config.ensure_auth(config=config)
                snapshot, zip_filepath = download_snapshot(auth, config)
                # Get the current timestamp
                timestamp = datetime.now().strftime("%Y%m%d%H%M%S")
                # Define the new directory name
                new_dir_name = config.work_dir / f"git-backup-{timestamp}"
                # Rename the directory
                shutil.move(".git", new_dir_name)
                # unzip the snapshot
                with zipfile.ZipFile(zip_filepath, "r") as zip_ref:
                    zip_ref.extractall(".")
            click.secho(
                f"Pulled {snapshot['file_name']}", fg=INFO_COLOR, file=sys.stderr
            )

        manifest_files, dvc_objects = manifest(config.gen3.project_id)
        if remote == "gen3":
            # download the files
            with Halo(
                text="Pulling from gen3",
                spinner="line",
                placement="right",
                color="white",
            ):
                object_ids = [
                    {"object_id": _.object_id} for _ in dvc_objects
                ]  # if not _.out.source_url
                current_time = datetime.now().strftime(
                    "%Y%m%d%H%M%S"
                )  # Format datetime as you need
                manifest_file = (
                    pathlib.Path(config.work_dir) / f"manifest-{current_time}.json"
                )
                with open(manifest_file, "w") as fp:
                    json.dump(object_ids, fp)
            cmd = f"gen3-client download-multiple --no-prompt --profile {config.gen3.profile}  --manifest {manifest_file} --numparallel {worker_count}"
            print(cmd)
            run_command(cmd, no_capture=True)
        elif remote == "s3":
            with Halo(
                text="Pulling from s3", spinner="line", placement="right", color="white"
            ):
                if not auth:
                    auth = gen3_tracker.config.ensure_auth(config=config)
                results = ls(
                    config, metadata={"project_id": config.gen3.project_id}, auth=auth
                )
                object_ids = [_.object_id for _ in dvc_objects]
            for _ in results["records"]:
                if _["did"] in object_ids:
                    print("aws s3 cp ", _["urls"][0], _["file_name"])
        elif remote == "ln":
            for _ in dvc_objects:
                print(f"ln -s {_.out.realpath} {_.out.path}")
        elif remote == "scp":
            for _ in dvc_objects:
                print(f"scp USER@HOST:{_.out.realpath} {_.out.path}")

        else:
            raise NotImplementedError(f"Remote {remote} not supported.")

    except Exception as e:
        click.secho(str(e), fg=ERROR_COLOR, file=sys.stderr)
        if config.debug:
            raise


@cli.command()
@click.argument(
    "project_id",
    default=None,
    required=False,
    envvar=f"{gen3_tracker.ENV_VARIABLE_PREFIX}PROJECT_ID",
    metavar="PROJECT_ID",
)
@click.pass_obj
def clone(config, project_id):
    """Clone a repository into a new directory"""
    try:
        config.gen3.project_id = project_id
        assert not pathlib.Path(
            project_id
        ).exists(), f"{project_id} already exists.  Please remove it first."
        os.mkdir(project_id)
        os.chdir(project_id)
        with Halo(text="Cloning", spinner="line", placement="right", color="white"):
            auth = gen3_tracker.config.ensure_auth(config=config)
            snapshot, zip_filepath = download_snapshot(auth, config)
            assert not pathlib.Path(
                ".git"
            ).exists(), "A git repository already exists.  Please remove it, or move to another directory first."
            # unzip
            with zipfile.ZipFile(zip_filepath, "r") as zip_ref:
                zip_ref.extractall(".")

            # if we just unzipped a .git these directories will exist
            expected_dirs = [".git", "META", "MANIFEST"]
            if not all([pathlib.Path(_).exists() for _ in expected_dirs]):
                # if not, we have downloaded a legacy SNAPSHOT.zip, so lets migrate the data to the expected drirectories
                click.secho(
                    f"{expected_dirs} not found after downloading {snapshot['file_name']} processing legacy snapshot",
                    fg=INFO_COLOR,
                    file=sys.stderr,
                )
                # legacy - was this a *SNAPSHOT.zip?
                meta_files = pathlib.Path("studies") / config.gen3.project
                # legacy - was this a *meta.zip?
                if not meta_files.exists():
                    meta_files = pathlib.Path(".")
                # create local directories and git
                [_ for _ in config_init(config, project_id)]
                ensure_git_repo(config=config)
                # move ndjson from studies to META
                for _ in meta_files.glob("*.ndjson"):
                    shutil.move(_, "META/")
                # add to git
                run_command("git add META/*.*")
                # migrate DocumentReferences to MANIFEST
                references = meta_index()
                manifest_files = []
                for _ in read_ndjson_file("META/DocumentReference.ndjson"):
                    document_reference = DocumentReference.parse_obj(_)
                    dvc_object = DVC.from_document_reference(
                        config, document_reference, references
                    )
                    manifest_files.append(
                        write_dvc_file(
                            yaml_data=dvc_object.model_dump(),
                            target=dvc_object.out.path,
                        )
                    )

                # Get the current time in seconds since the epoch
                current_time = time.time()
                # Update the access and modification times of the file
                os.utime("META/DocumentReference.ndjson", (current_time, current_time))

                run_command("git add MANIFEST/")
                run_command(
                    'git commit -m "migrated from legacy" MANIFEST/ META/ .gitignore'
                )
                shutil.move(zip_filepath, config.work_dir / zip_filepath.name)

        click.secho(f"Cloned {snapshot['file_name']}", fg=INFO_COLOR, file=sys.stderr)
        run_command("git status", no_capture=True)

    except Exception as e:
        click.secho(str(e), fg=ERROR_COLOR, file=sys.stderr)
        if config.debug:
            raise


def download_snapshot(auth, config):
    """Download the latest snapshot."""
    from gen3_tracker.git.cloner import find_latest_snapshot

    snapshot = find_latest_snapshot(auth, config)

    gen3_file = Gen3File(auth)
    pathlib.Path(snapshot["file_name"]).parent.mkdir(exist_ok=True, parents=True)
    ok = gen3_file.download_single(snapshot["did"], ".")
    assert ok, f"Failed to download {snapshot['did']}"

    zip_filepath = pathlib.Path(snapshot["file_name"])
    assert zip_filepath.exists(), f"Failed to download {snapshot['did']}"
    return snapshot, zip_filepath


def file_name_or_guid(config, object_id) -> (str, pathlib.Path):
    """Check if the object_id is a file name or a GUID."""
    guid_pattern = re.compile(
        r"^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$"
    )
    path = None
    if not guid_pattern.match(object_id):
        if not is_url(object_id):
            path = pathlib.Path("MANIFEST") / (object_id + ".dvc")
        else:
            path = pathlib.Path("MANIFEST") / (url_path(object_id) + ".dvc")

        if path.exists():
            dvc_object = next(iter(dvc_data([str(path)])), None)
            assert dvc_object, f"{object_id} is not a valid GUID or DVC file."
            dvc_object.project_id = config.gen3.project_id
            object_id = dvc_object.object_id
        else:
            raise ValueError(
                f"{object_id} was not found in the MANIFEST and does not appear to be an object identifier (GUID)."
            )
    else:
        committed_files, dvc_objects = manifest(config.gen3.project_id)
        dvc_objects = [
            dvc_object
            for dvc_object in dvc_objects
            if dvc_object.object_id == object_id
        ]
        assert dvc_objects, f"{object_id} not found in MANIFEST."
        path = pathlib.Path("MANIFEST") / (dvc_objects[0].out.path + ".dvc")

    assert guid_pattern.match(object_id), f"{object_id} was not found in MANIFEST."
    return object_id, path


@cli.command("ls")
@click.option(
    "--long",
    "-l",
    "long_flag",
    default=False,
    is_flag=True,
    help="Long listing format.",
    show_default=True,
)
@click.argument("target", default=None, required=False)
@click.pass_obj
def ls_cli(config: Config, long_flag: bool, target: str):
    """List files in the repository.
    \b
    TARGET wild card match of guid, path or hash.
    """
    try:

        with Halo(
            text="Pulling file list", spinner="line", placement="right", color="white"
        ):
            auth = gen3_tracker.config.ensure_auth(config=config)
            results = ls(
                config, metadata={"project_id": config.gen3.project_id}, auth=auth
            )
            indexd_records = results["records"]
            committed_files, dvc_objects = manifest(config.gen3.project_id)
            # list all data files
            dvc_objects = {_.object_id: _ for _ in dvc_objects}

            def _dvc_meta(dvc_object, full=False) -> dict:
                if not dvc_object:
                    return {}
                _ = {}
                if not full and dvc_object.meta:
                    for k, v in dvc_object.meta.model_dump(exclude_none=True).items():
                        if v:
                            _[k] = v
                else:
                    _ = dvc_object.model_dump(exclude_none=True)
                _["object_id"] = dvc_object.object_id
                return _

            if not long_flag:
                indexd_records = [
                    {
                        "did": _["did"],
                        "file_name": _["file_name"],
                        "indexd_created_date": _["created_date"],
                        "meta": _dvc_meta(dvc_objects.get(_["did"], None)),
                        "urls": _["urls"],
                    }
                    for _ in indexd_records
                ]

        bucket_ids = {_["did"] for _ in indexd_records}

        uncommitted = pathlib.Path("MANIFEST").glob("**/*.dvc")
        uncommitted = [str(_) for _ in uncommitted]
        uncommitted = [str(_) for _ in uncommitted if _ not in committed_files]
        uncommitted = [_.model_dump(exclude_none=True) for _ in dvc_data(uncommitted)]

        _ = {
            "bucket": indexd_records,
            "committed": [
                _dvc_meta(v, full=True)
                for k, v in dvc_objects.items()
                if k not in bucket_ids
            ],
            "uncommitted": uncommitted,
        }

        if target:
            # Escape special characters and replace wildcard '*' with '.*' for regex pattern
            pattern = re.escape(target).replace("\\*", ".*")
            filtered = {
                "bucket": filter_dicts(_.get("bucket", []), pattern),
                "committed": filter_dicts(_.get("committed", []), pattern),
                "uncommitted": filter_dicts(_.get("uncommitted", []), pattern),
            }
            _ = filtered

        if config.output.format == "json":
            print(json.dumps(_, indent=2))
        else:
            yaml.dump(_, sys.stdout, default_flow_style=False)

    except Exception as e:
        click.secho(str(e), fg=ERROR_COLOR, file=sys.stderr)
        if config.debug:
            raise


@cli.command()
@click.argument("object_id", metavar="<name>")
@click.pass_obj
def rm(config: Config, object_id: str):
    """Remove a single file from the server index, and MANIFEST. Does not alter META.
    \b
    <name> is a GUID or a data file name.
    """
    try:

        with Halo(text="Searching", spinner="line", placement="right", color="white"):
            object_id, path = file_name_or_guid(config, object_id)

        with Halo(
            text="Deleting from server",
            spinner="line",
            placement="right",
            color="white",
        ):
            auth = gen3_tracker.config.ensure_auth(config=config)
            index = Gen3Index(auth)
            result = index.delete_record(object_id)
        if not result:
            if not path:
                path = ""
            click.secho(
                f"Failed to delete {object_id} from server. {path}",
                fg=ERROR_COLOR,
                file=sys.stderr,
            )
        else:
            click.secho(
                f"Deleted {object_id} from server. {path}",
                fg=INFO_COLOR,
                file=sys.stderr,
            )

        with Halo(text="Scanning", spinner="line", placement="right", color="white"):
            committed_files, dvc_objects = manifest(config.gen3.project_id)
            dvc_objects = [
                dvc_object
                for dvc_object in dvc_objects
                if dvc_object.object_id == object_id
            ]
            assert dvc_objects, f"{object_id} not found in MANIFEST."
            dvc_object = dvc_objects[0]
            path = pathlib.Path("MANIFEST") / (dvc_object.out.path + ".dvc")
            assert path.exists(), f"{path} not found"
            path.unlink()
        click.secho(
            f"Deleted {path} from MANIFEST. Please adjust META resources",
            fg=INFO_COLOR,
            file=sys.stderr,
        )

    except Exception as e:
        click.secho(str(e), fg=ERROR_COLOR, file=sys.stderr)
        if config.debug:
            raise


@cli.command(name="ping")
@click.pass_obj
def ping(config: Config):
    """Verify gen3-client and test connectivity."""
    with CLIOutput(config=config) as output:
        msgs = []
        ok = True
        cmd = "gen3-client --version".split()
        gen3_client_installed = subprocess.run(cmd, capture_output=True)
        if gen3_client_installed.returncode != 0:
            msgs.append("gen3-client not installed")
            ok = False

        gen_client_ini_file = gen3_tracker.config.gen_client_ini_path()
        auth = None
        if not gen_client_ini_file.exists():
            msgs.append("not configured")
            ok = False
        else:
            try:
                assert config.gen3.profile, "No profile found"
                auth = ensure_auth(config=config, validate=True)
                assert auth, "Authentication failed"
                msgs.append(f"Connected using profile:{config.gen3.profile}")
            except (AssertionError, ValueError) as e:
                msgs.append(str(e))
                ok = False
            except Gen3AuthError as e:
                msg = str(e).split(":")[0]
                msgs.append(msg)
                msg2 = str(e).split('<p class="introduction">')[-1]
                msg2 = msg2.split("</p>")[0]
                msgs.append(msg2)
                ok = False

        if ok:
            _ = "Configuration OK: "
        else:
            _ = "Configuration ERROR: "
            output.exit_code = 1

        _ = {"msg": _ + ", ".join(msgs)}
        if auth:
            _["endpoint"] = auth.endpoint
            user_info = auth.curl("/user/user").json()
            _["username"] = user_info["username"]
            buckets = get_buckets(config=config)
            bucket_info = {}
            program_info = defaultdict(list)
            for k, v in buckets["S3_BUCKETS"].items():
                bucket_info[k] = {}
                if "programs" not in v:
                    bucket_info[k] = "No `programs` found"
                    click.secho(
                        f"WARNING: No `programs` found for bucket {k}",
                        fg=INFO_COLOR,
                        file=sys.stderr,
                    )
                    continue
                bucket_info[k] = ",".join(v["programs"])
                for program in v["programs"]:
                    program_info[program].append(k)
            _["bucket_programs"] = bucket_info

            for k, v in program_info.items():
                if len(v) > 1:
                    click.secho(
                        f"WARNING: {k} is in multiple buckets: {', '.join(v)}",
                        fg=INFO_COLOR,
                        file=sys.stderr,
                    )

            assert "authz" in user_info, "No authz found"
            authz_info = defaultdict(dict)
            for k, v in user_info["authz"].items():
                authz_info[k] = ",".join(set([_["method"] for _ in v]))
            _["your_access"] = dict(authz_info)

        output.update(_)


# @cli.command()
# @click.argument('targets', nargs=-1)
# @click.pass_obj
# def diff(config: Config, targets):
#     """Show details of changed files."""
#     try:
#         click.secho(DIFF_MESSAGE, fg=INFO_COLOR, file=sys.stderr)
#         changes = git_status(config.debug)
#
#         if not targets:
#             targets = []
#             for k in ['committed', 'uncommitted']:
#                 if k in changes:
#                     targets.extend(changes[k]['modified'])
#                     click.secho(f"Targets not specified, comparing {k} changes", fg=INFO_COLOR, file=sys.stderr)
#
#         if not targets:
#             click.secho("No changes found", fg=INFO_COLOR, file=sys.stderr)
#             return
#
#         # check committed changes
#         for _ in targets:
#             if _.endswith('.dvc'):
#                 continue
#             if _ in changes.get('committed', {'modified': []})['modified']:
#                 click.secho(f'{_} is in the committed changes.', fg=INFO_COLOR, file=sys.stderr)
#                 # read the git diff, get the hashes of the changed files
#                 results = run_command(f'git diff {_}.dvc', config.debug)
#                 hashes = [_.replace('md5:', '') for _ in results.stdout.split('\n') if 'md5:' in _]
#                 old = hashes[0].split()[-1]
#                 new = hashes[1].split()[-1]
#                 # there should be a change in the hash
#                 assert old != new, f'{_} has not changed.'
#                 # get the file paths from the dvc cache
#                 # see https://dvc.org/doc/user-guide/project-structure/internal-files#structure-of-the-cache-directory
#                 dvc_cache_path = pathlib.Path('.dvc/cache/files/md5').resolve()
#                 old_file = dvc_cache_path / old[:2] / old[2:]
#                 new_file = dvc_cache_path / new[:2] / new[2:]
#                 # run diff on the files
#                 results = run_command(f'diff -u {old_file} {new_file}', config.debug, raise_on_err=False)
#                 click.secho(_, fg=INFO_COLOR, file=sys.stderr)
#                 print(results.stdout)
#             else:
#                 # click.secho(f'{_} is in the uncommitted changes.', fg=INFO_COLOR, file=sys.stderr)
#                 # get the hash of the original file
#                 changes = dvc_diff_head(config.debug)
#                 modified = changes.get('modified', [])
#                 for modified_file in modified:
#                     if modified_file['path'] == _:
#                         old = modified_file['hash']['old']
#                         dvc_cache_path = pathlib.Path('.dvc/cache/files/md5').resolve()
#                         old_file = dvc_cache_path / old[:2] / old[2:]
#                         new_file = modified_file['path']
#                         # run diff on the files
#                         print(f'diff -u {old_file} {new_file}')
#                         results = run_command(f'diff -u {old_file} {new_file}', config.debug, raise_on_err=False)
#                         click.secho(_, fg=INFO_COLOR, file=sys.stderr)
#                         print(results.stdout)
#                         break
#
#     except Exception as e:
#         click.secho(str(e), fg=ERROR_COLOR, file=sys.stderr)
#         if config.debug:
#             raise
#
#
# @cli.command()
# @click.argument('branch')
# @click.pass_obj
# def checkout(ctx: click.Context, branch: str):
#     """Checkout a particular branch or commit."""
#     try:
#         click.secho(f"Checking out {branch}", fg=INFO_COLOR, file=sys.stderr)
#         run_command(f'git checkout {branch}', config.debug, no_capture=True)
#         run_command(f'dvc checkout', config.debug, no_capture=True)
#     except Exception as e:
#         click.secho(str(e), fg=ERROR_COLOR, file=sys.stderr)
#         if config.debug:
#             raise



from setuptools import setup, find_packages
# import os
# print(os.getcwd())

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

with open('README.md', 'r') as f:
    long_description = f.read()


setup(
    name='gen3_tracker',
    version='0.0.7rc11',
    description='A CLI for adding version control to Gen3 data submission projects.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='walsbr',
    author_email='walsbr@ohsu.edu',
    url='https://github.com/ACED-IDP/gen3_util',
    packages=find_packages(exclude=['tests', 'tests.*']),
    install_requires=requirements,
    include_package_data=True,
    package_data={  # Optional
        '': ['*.yaml'],
    },
    extras_require={
        'dtale': ['dtale'],
    },
    entry_points={
        'console_scripts': [
            'g3t=gen3_tracker.cli:cli',
        ],
    },
)



import requests
from gen3.auth import Gen3Auth
from requests import HTTPError

from gen3_tracker.config import ensure_auth, Config


def _ensure_auth(auth, config):
    """Create auth from config, if we don't have one already."""
    if not auth:
        assert config
        auth = ensure_auth(config=config)
    return auth


def get_requests(config: Config = None, auth: Gen3Auth = None, mine: bool = False, active: bool = False, username: str = None) -> dict:
    """Fetch information about the user.
    Parameters
    ----------
    config : Config
        The config object.
    auth : Gen3Auth
        The auth object.
    mine : bool
        If True, return requests for the current user.
    active : bool
        If True, return only active (non-final requests) requests. see https://github.com/uc-cdis/requestor/blob/master/src/requestor/config-default.yaml#L63
    username : str
        If provided, return requests for this user.
    """
    auth = _ensure_auth(auth, config)
    if mine:
        # returns a list of dicts
        # https://github.com/uc-cdis/requestor/blob/master/src/requestor/routes/query.py#L200
        url = '/requestor/request/user'
        parms = []
        if active:
            parms.append("active")
        if len(parms) > 0:
            url = url + "?" + "&".join(parms)
        return auth.curl(url).json()
    else:
        # returns a list of dicts
        # https://github.com/uc-cdis/requestor/blob/master/src/requestor/routes/query.py#L158
        url = '/requestor/request'
        parms = []
        if username:
            parms.append(f"username={username}")
        if active:
            parms.append("active")
        if len(parms) > 0:
            url = url + "?" + "&".join(parms)
        return auth.curl(url).json()


def get_request(config: Config = None, auth: Gen3Auth = None, request_id: str = None):
    """Get a specific request"""
    assert request_id, "required"
    auth = _ensure_auth(auth, config)
    # returns a dict
    # https://github.com/uc-cdis/requestor/blob/master/src/requestor/routes/query.py#L235
    return auth.curl(f'/requestor/request/{request_id}').json()


def create_request(config: Config = None, auth: Gen3Auth = None, request: dict = None, revoke: bin = False):
    """Create a specific request"""
    auth = _ensure_auth(auth, config)
    one_of = ['policy_id', 'resource_paths', 'resource_path']
    assert any([k in request for k in one_of]), (f"one of {one_of} required", request)

    url = auth.endpoint + "/" + 'requestor/request'
    if revoke:
        url = url + "?revoke"

    response = requests.post(
        url, json=request, auth=auth
    )

    try:
        response.raise_for_status()
    except HTTPError as e:
        print("error:", e)
        print("request:", request)
        print("response.text", response.text)
        raise e

    return response.json()


def update_request(config: Config = None, auth: Gen3Auth = None, request_id: str = None, status: str = None):
    """Update a specific request"""
    assert request_id, "required"
    assert status, "required"
    auth = _ensure_auth(auth, config)
    request = {'status': status}

    response = requests.put(
        auth.endpoint + "/" + f'requestor/request/{request_id}', json=request, auth=auth
    )
    response.raise_for_status()
    return response.json()



import glob
import pathlib
import sys
from datetime import datetime
from urllib.parse import urlparse, ParseResult

import click
import dateutil
import pytz
import yaml
from dateutil.tz import tzutc

from gen3_tracker import Config
from gen3_tracker.common import ACCEPTABLE_HASHES
from gen3_tracker.git import DVC


def is_valid_url(target):
    try:
        result = urlparse(target)
        return all([result.scheme, result.netloc])
    except ValueError:
        return False


# handle http, https
def get_http_cloud_storage_path(url):
    parsed_url = urlparse(url)
    if parsed_url.scheme not in ['http', 'https']:
        raise ValueError(f'Invalid scheme {parsed_url.scheme} for cloud storage URL')

    # The path component of the parsed URL starts with a '/', so we remove it with [1:]
    # For MinIO, the path includes the bucket name, so we need to split on '/' and take everything after the first element
    return '/'.join(parsed_url.path.split('/')[2:])


def get_http_bucket_from_url(url):
    parsed_url = urlparse(url)
    if parsed_url.scheme not in ['http', 'https']:
        raise ValueError(f'Invalid scheme {parsed_url.scheme} for cloud storage URL')

    # For MinIO, the bucket name is the first element in the path
    return parsed_url.path.split('/')[1]


# handle s3, gs, azure
def get_cloud_storage_path(url) -> ParseResult:
    """breaks the URL into components, which can then be used to derive the file path."""
    parsed_url: ParseResult = urlparse(url)
    if parsed_url.scheme not in ['s3', 'gs', 'azure']:
        raise ValueError(f'Invalid scheme {parsed_url.scheme} for cloud storage URL')

    # The path component of the parsed URL starts with a '/', so we remove it with [1:]
    return parsed_url


def add_url(ctx, target) -> tuple[list[pathlib.Path], list[str]]:
    from gen3_tracker.common import ACCEPTABLE_HASHES, INFO_COLOR

    """Add a url to the repository. If the file is already in the repository, it will be updated."""
    assert is_valid_url(target), f'{target} is not a valid url.'
    url = target
    from gen3_tracker.git import git_files

    files_already_in_repo = git_files()
    updates = []  # only updates
    all_changed_files = []  # new and updates
    config: Config = ctx.obj

    target = url_path(url)

    if f"MANIFEST/{target}.dvc" in files_already_in_repo:
        click.secho(f'{target} is already in the repository. updating.', fg=INFO_COLOR, file=sys.stderr)
        # flag the file for update
        updates.append(target)

    # process args to DVC
    _args = ctx.args
    metadata = {'meta': dict(map(lambda i: (_args[i].replace('--', ''), _args[i + 1]), range(len(_args) - 1)[::2]))}
    if '--no-bucket' in _args or '--no_bucket' in _args:
        metadata['meta']['no_bucket'] = True

    required_keys = ['size', 'modified']
    required_keys_msg = []
    for k in required_keys:
        if k not in metadata['meta']:
            required_keys_msg.append(f'--{k} is required for urls.')
    assert not required_keys_msg, f'{", ".join(required_keys_msg)}'

    # check content of parameters
    assert metadata['meta']['size'].isdigit(), 'size should be an integer.'
    metadata['meta']['size'] = int(metadata['meta']['size'])

    # assuming metadata['meta']['modified'] is a string representing a date
    date_str = metadata['meta']['modified']
    # parse the string into a datetime object
    date_obj = dateutil.parser.parse(date_str)
    # if the date string doesn't have a timezone, you can add one
    if date_obj.tzinfo is None:
        date_obj = date_obj.replace(tzinfo=tzutc())
    # convert the datetime object back into an ISO formatted string with timezone
    metadata['meta']['modified'] = date_obj.isoformat()

    for k in ACCEPTABLE_HASHES.keys():
        if k in metadata['meta']:
            hash_type = k
            hash_value = metadata['meta'][k]
            assert hash_value, f'{hash_type} should be provided.'
            hash_regex = ACCEPTABLE_HASHES[hash_type]
            assert hash_regex(hash_value), f'{hash_value} is not a valid {hash_type} hash.'
            if 'hash' in metadata['meta']:
                assert metadata['meta']['hash'] == hash_type, f'--hash {hash_type} does not match the provided hash type {metadata["meta"]["hash"]}.'
            else:
                metadata['meta']['hash'] = hash_type
            break

    # create reference to the file
    metadata['project_id'] = config.gen3.project_id
    yaml_data = create_dvc_for_url(metadata, target=target)
    yaml_data.update(metadata)
    yaml_data['outs'][0]['source_url'] = url
    # delete file keys from metadata
    for k in required_keys:
        del yaml_data['meta'][k]
    for k in ACCEPTABLE_HASHES.keys():
        if k in metadata['meta']:
            del yaml_data['meta'][k]

    # write the dvc file
    dvc_file = write_dvc_file(target, yaml_data)

    all_changed_files = [dvc_file]
    return all_changed_files, updates


def url_path(url):
    """Breaks the URL into components, which can then be used to derive the MANIFEST file path."""
    # The path component of the parsed URL starts with a '/', so we remove it with [1:]
    if not url.startswith('http'):
        cloud_storage_parts = get_cloud_storage_path(url)
        path = cloud_storage_parts.path[1:]
        bucket = cloud_storage_parts.netloc
    else:
        path = get_http_cloud_storage_path(url)
        bucket = get_http_bucket_from_url(url)
    target = f"{bucket}/{path}"
    return target


def add_file(ctx, target) -> tuple[list[pathlib.Path], list[str]]:
    from gen3_tracker.common import INFO_COLOR
    config: Config = ctx.obj

    """Add a real file to the repository. Expand wildcards. If the file is already in the repository, it will be updated."""
    from gen3_tracker.git import git_files

    targets = glob.glob(target, recursive=True)
    if not targets:
        # not a wildcard, check if it is a file
        # assert pathlib.Path(target).exists(), f'{pathlib.Path(target).resolve()} does not exist.'
        targets = [target]

    files_already_in_repo = git_files()
    updates = []  # only updates
    all_changed_files = []  # new and updates
    for target in targets:

        # check args & dependencies
        # target should be a file that exists and is relative to the project root
        if target.startswith('MANIFEST') or target.startswith('META'):
            suggested_name = target.replace('MANIFEST/', '').replace('META/', '').replace('.dvc', '')
            click.secho(f'{target} starts with a reserved name. Perhaps you meant {suggested_name}?', fg=INFO_COLOR,
                        file=sys.stderr)
            continue

        if f"MANIFEST/{target}.dvc" in files_already_in_repo:
            click.secho(f'{target} is already in the repository. updating.', fg=INFO_COLOR, file=sys.stderr)
            #
            # flag the file for update
            #
            updates.append(target)

        target_path = pathlib.Path(target)
        # assert not target_path.is_dir(), 'Adding directories are not supported. Please supply a file.'
        if target_path.is_dir():
            continue

        # final checks
        # assert target_path.resolve().exists(), f'{pathlib.Path(target).resolve()} does not exist.'
        assert target_path.resolve().is_relative_to(pathlib.Path.cwd()), 'Target should be relative to the project root.'

        # create reference to the file
        # convert --arguments to metadata
        _args = ctx.args
        metadata = {'meta': dict(map(lambda i: (_args[i].replace('--', ''), _args[i + 1]), range(len(_args) - 1)[::2]))}
        if '--no-bucket' in _args or '--no_bucket' in _args:
            metadata['meta']['no_bucket'] = True

        if 'hash' in metadata['meta']:
            hash_type = metadata['meta']['hash']
            assert hash_type in ACCEPTABLE_HASHES.keys(), f'hash should be one of {", ".join(ACCEPTABLE_HASHES.keys())}'
            hash_value = metadata['meta'][hash_type]
            assert hash_value, f'{hash_type} should be provided.'
            hash_regex = ACCEPTABLE_HASHES[hash_type]
            assert hash_regex(hash_value), f'{hash_value} is not a valid {hash_type} hash.'

        yaml_data = create_dvc(metadata, target_path)
        yaml_data.update(metadata)
        yaml_data['project_id'] = config.gen3.project_id
        _ = DVC(**yaml_data).model_dump()
        assert _['outs'][0]['object_id'], 'object_id should be in dvc created from files.'
        dvc_file = write_dvc_file(target, _)

        # add the target root to the gitignore
        git_ignore_path = pathlib.Path.cwd() / '.gitignore'
        all_changed_files.append(dvc_file)
        ignores = []
        if not git_ignore_path.exists():
            all_changed_files.append(git_ignore_path)
        else:
            with open(git_ignore_path) as f:
                ignores = f.readlines()

        target_root = f'/{target_path.parts[0]}\n'
        if target_root not in ignores:
            with open(git_ignore_path, 'a+') as f:
                f.write(target_root)

    return all_changed_files, updates


def write_dvc_file(target, yaml_data):
    # write the dvc file
    manifest_path = pathlib.Path('MANIFEST')
    # meta_path = pathlib.Path('META')
    manifest_target_path = manifest_path / target
    manifest_target_path.parent.mkdir(parents=True, exist_ok=True)
    dvc_file = manifest_target_path.parent / (manifest_target_path.name + ".dvc")
    if dvc_file.exists():
        with open(dvc_file) as f:
            existing_yaml_data = yaml.load(f, Loader=yaml.SafeLoader)
            # don't overwrite existing metadata with empty metadata
            for k, v in existing_yaml_data['meta'].items():
                if v is None:
                    continue
                if k in yaml_data['meta'] and yaml_data['meta'][k] is not None:
                    continue
                yaml_data['meta'][k] = v
    with open(dvc_file, 'w') as yaml_file:
        yaml.dump(yaml_data, yaml_file, default_flow_style=False)
    return dvc_file


def create_dvc(metadata: dict, target_path: pathlib.Path) -> dict:
    from gen3_tracker.common import ACCEPTABLE_HASHES

    """Create a dvc file for a file in the repository."""
    from gen3_tracker.git import calculate_hash, get_mime_type

    target = str(target_path)

    info = {
        'size': metadata['meta'].get('size', None) or target_path.stat().st_size,
        'path': target,
        'mime': metadata['meta'].get('mime', get_mime_type(target)),
        'modified': metadata['meta'].get('modified', None) or datetime.fromtimestamp(target_path.stat().st_mtime, tz=pytz.utc).isoformat(),
        'realpath': str(target_path.resolve()),
        'is_symlink': target_path.is_symlink()
    }
    if 'realpath' in metadata['meta']:
        info['realpath'] = metadata['meta']['realpath']
        del metadata['meta']['realpath']
    # did they provide a hash?

    for k in ACCEPTABLE_HASHES.keys():
        if k in metadata['meta']:
            info[k] = metadata['meta'][k]
            info['hash'] = k
            del metadata['meta'][k]
            if 'hash' in metadata['meta']:
                del metadata['meta']['hash']
    # no?, use md5
    if 'hash' not in info:
        info['hash'] = 'md5'
        info['md5'] = calculate_hash('md5', target)
    # we follow this convention for the dvc file
    # see https://dvc.org/doc/user-guide/project-structure/dvc-files#dvc-files
    yaml_data = {
        'outs': [info]
    }
    return yaml_data


def create_dvc_for_url(metadata: dict, target: str) -> dict:
    """Create a dvc file for a url in the repository."""
    from gen3_tracker.git import get_mime_type
    from gen3_tracker.common import ACCEPTABLE_HASHES

    info = {
        'size': metadata['meta'].get('size', None),
        'path': target,
        'mime': metadata['meta'].get('mime', get_mime_type(target)),
        'modified': metadata['meta'].get('modified', None),
        'realpath': None,
        'is_symlink': False
    }
    # did they provide a hash?
    for k in ACCEPTABLE_HASHES.keys():
        if k in metadata['meta']:
            info[k] = metadata['meta'][k]
            info['hash'] = k
    # we follow this convention for the dvc file
    # see https://dvc.org/doc/user-guide/project-structure/dvc-files#dvc-files
    yaml_data = {
        'project_id': metadata['project_id'],
        'outs': [info]
    }
    _ = DVC(**yaml_data).model_dump()
    assert _['outs'][0]['object_id'], f'object_id should be in dvc created for url. {metadata}'
    return _
